"""Process club records."""
import csv
import datetime
import os
import re
import time
from pathlib import Path
from shutil import copyfile

from scm_helper.config import (
    C_AGE_EOY,
    C_ALL_AGES,
    C_OVERALL_FASTEST,
    C_RECORDS,
    C_RELAY,
    C_SE_ONLY,
    C_VERIFY,
    CONFIG_DIR,
    FILE_READ,
    FILE_WRITE,
    PRINT_DATE_FORMAT,
    RECORDS_DIR,
    SCM_CSV_DATE_FORMAT,
    get_config,
)
from scm_helper.issue import debug
from scm_helper.notify import notify

STROKES = {
    "Free": "Freestyle",
    "Back": "Backstroke",
    "Breast": "Breaststroke",
    "Fly": "Butterfly",
    "Medley": "Individual Medley",
}

RELAY_STROKES = {
    "Free": "Freestyle",
    "Medley": "Medley",
}

DISTANCE = [
    "50m",
    "100m",
    "200m",
    "400m",
    "800m",
    "1500m",
]

STROKE_DISTANCE = {
    "Free": ["50m", "100m", "200m", "400m", "800m", "1500m"],
    "Back": ["50m", "100m", "200m"],
    "Breast": ["50m", "100m", "200m"],
    "Fly": ["50m", "100m", "200m"],
    "Medley": ["100m", "200m", "400m"],
}

RELAY_DISTANCE = {"200": 200, "400": 400, "800": 800}
PRINT_DISTANCE = {"200": "4 x 50", "400": "4 x 100", "800": "4 x 200"}
OVERALL = "Overall"

AGES = {
    OVERALL: 1,
    "6-7": 0,
    "8-9": 0,
    "10-11": 0,
    "12-13": 0,
    "14-15": 0,
    "16-17": 0,
    "18-24": 0,
    "25-29": 0,
    "30-34": 0,
    "35-39": 0,
    "40-44": 0,
    "45-49": 0,
    "50-54": 0,
    "55-59": 0,
    "60-64": 0,
    "65-69": 0,
    "70-74": 0,
    "75-79": 0,
    "80-84": 0,
    "85-89": 0,
    "90-94": 0,
    "95-99": 0,
}

ALL_AGES = {
    OVERALL: 1,
    "6": 0,
    "7": 0,
    "8": 0,
    "9": 0,
    "10": 0,
    "11": 0,
    "12": 0,
    "13": 0,
    "14": 0,
    "15": 0,
    "16": 0,
    "17": 0,
    "18": 0,
    "19-24": 0,
    "25-29": 0,
    "30-34": 0,
    "35-39": 0,
    "40-44": 0,
    "45-49": 0,
    "50-54": 0,
    "55-59": 0,
    "60-64": 0,
    "65-69": 0,
    "70-74": 0,
    "75-79": 0,
    "80-84": 0,
    "85-89": 0,
    "90-94": 0,
    "95-99": 0,
}

RELAY_AGES = {
    "72": "72+",
    "100": "100-119",
    "120": "120-159",
    "160": "160-199",
    "200": "200-239",
    "240": "240-279",
    "280": "280+",
}

GENDER = {"M": "Male", "F": "Female"}
RELAY_GENDER = {"M": "Male", "F": "Female", "Mixed": "Mixed"}

COURSE = {"25": "SC", "50": "LC"}

R_SWIMMER1 = "swimmer 1"
R_SWIMMER2 = "swimmer 2"
R_SWIMMER3 = "swimmer 3"
R_SWIMMER4 = "swimmer 4"
R_TIME = "time"
R_LOCATION = "location"
R_DATE = "date"

TAG_INNER_EVEN = "name=record-inner-even"
TAG_INNER_ODD = "name=record-inner-odd"

S_EVENT = "event"
S_ASA = "asa"
S_NAME = "name"
S_TIMESTR = "time"
S_FTIME = "ftime"
S_LOCATION = "location"
S_DATE = "date"

F_BASELINE = "records.csv"
F_RECORDS = "records.html"
F_RELAY_BASELINE = "relay_records.csv"
F_RELAY_RECORDS = "relay_records.html"

HEADER = "_header.txt"
FOOTER = """<p>
Records page created using
<a href="https://github.com/ColinRobbins/scm-helper">SCM Helper</a>.
</p>
"""

INNER_WRAP = """
<div class=divTableRow XX_TAG_XX>
<div class=divTableCell>
Age: XX_AGE_XX:
</div>
<div class=divTableCell>
XX_OPT_XX
</div>
</div>
"""

AGE_WRAP = """
<h3>XX_DIST_XX:</h3>
<div class=divTable>
<div class=divTableBody>
XX_AGE_XX
</div>
</div>
"""

DIST_WRAP = """
<div class="divTable" name=\"record-XX_STROKE_XX\">
XX_DIST_XX
</div>
"""

GENDER_WRAP = """
<div name=\"record-XX_GENDER_XX\">
XX_O_GENDER_XX
</div>
"""

WRAP_TABLE_OPEN = " <div class=divTable><div class=divTableBody>\n"
WRAP_TABLE_CLOSE = " </div></div>\n"


class Records:
    """Read and process record files."""

    def __init__(self, scm):
        """Initialise."""
        self.scm = scm
        self.records = None
        self.relay = None

        try:
            home = str(Path.home())
            mydir = os.path.join(home, CONFIG_DIR, RECORDS_DIR)

            if os.path.exists(mydir) is False:
                os.mkdir(mydir)

            filename = os.path.join(mydir, F_BASELINE)
            if os.path.isfile(filename) is False:
                with open(filename, FILE_WRITE) as file:
                    file.write(DEFAULT_RECORDS)

            filename = os.path.splitext(filename)[0]
            filename += HEADER

            if os.path.isfile(filename) is False:
                with open(filename, FILE_WRITE) as file:
                    file.write(DEFAULT_HEADER)

        except EnvironmentError as error:
            notify(f"Cannot create default records:\n{error}\n")

    def read_baseline(self):
        """Read baseline."""

        home = str(Path.home())
        mydir = os.path.join(home, CONFIG_DIR, RECORDS_DIR)

        self.records = Record()
        filename = os.path.join(mydir, F_BASELINE)
        res = self.records.read_baseline(filename, self.scm)

        if res and get_config(self.scm, C_RECORDS, C_RELAY):
            self.relay = RelayRecord()
            filename = os.path.join(mydir, F_RELAY_BASELINE)
            res = self.relay.read_baseline(filename, self.scm)

        return res

    def read_newtimes(self, filename):
        """Read swimtimes."""

        times = SwimTimes(self.records)
        res = times.merge_times(filename, self.scm)
        if self.records.newrecords:
            for record in sorted(self.records.newrecords):
                notify(self.records.newrecords[record])

            self.records.write_records()

        del times
        return res

    def create_html(self):
        """Create HTML files for records."""

        home = str(Path.home())
        mydir = os.path.join(home, CONFIG_DIR, RECORDS_DIR)

        if get_config(self.scm, C_RECORDS, C_ALL_AGES):
            ages = ALL_AGES
        else:
            ages = AGES

        res = self.records.create_html(GENDER, STROKES, ages, False)
        filename = os.path.join(mydir, F_RECORDS)

        try:
            with open(filename, FILE_WRITE) as htmlfile:
                htmlfile.write(res)

        except EnvironmentError as error:
            notify(f"Cannot create HTML file: {filename}\n{error}\n")
            return False

        notify(f"Created {filename}...\n")

        if self.relay:
            res = self.relay.create_html(RELAY_GENDER, RELAY_STROKES, RELAY_AGES, True)
            filename = os.path.join(mydir, F_RELAY_RECORDS)

            try:
                with open(filename, FILE_WRITE) as htmlfile:
                    htmlfile.write(res)

            except EnvironmentError as error:
                notify(f"Cannot create HTML file: {filename}\n{error}\n")
                return False

            notify(f"Created {filename}...\n")

        return filename

    def delete(self):
        """Delete."""
        del self.records
        if self.relay:
            del self.relay


class SwimTimes:
    """Read SwimTimes, and merge into Records."""

    def __init__(self, records):
        """Initilaise Records handling."""
        self._filename = None
        self.scm = None
        self.records = records

    def merge_times(self, filename, scm):
        """Read Facebook file."""
        self._filename = filename
        self.scm = scm

        notify(f"Reading {filename}...\n")

        try:
            count = 0
            with open(filename, newline="", encoding="utf-8-sig") as csvfile:
                csv_reader = csv.DictReader(csvfile)

                for row in csv_reader:
                    count += 1

                    if (int(count / 1000)) == (count / 1000):
                        notify(f"{count} ")

                    if (int(count / 10000)) == (count / 10000):
                        notify("\n")

                    self.process_row(row, count)

            notify(f"\nRead {filename}...\n")
            return True

        except EnvironmentError as error:
            notify(f"Cannot open swim time file: {filename}\n{error}\n")
            return False

        except csv.Error as error:
            notify(f"Error in swim time file: {filename}\n{error}\n")
            return False

        notify("Unknown error\n")
        return False

    def process_row(self, row, count):
        """Process and merge a row into records."""

        # pylint: disable=too-many-locals
        # pylint: disable=too-many-return-statements
        # pylint: disable=too-many-branches
        # pylint: disable=too-many-statements

        if "Swimmer" not in row:
            if count == 1:
                notify("Is the header line missing in the CSV?\n")
            return

        swimmer = row["Swimmer"]
        asa = row["SE Number"]
        xdate = row["Date"]
        pool = row["Pool Size"]
        dist = row["Swim Distance"]
        stroke = row["Stroke"]
        timestr = row["Time"]
        relay = row["Relay"]
        location = row["Location"]
        gender = row["Gender"]
        gala = row["Gala"]

        swimage = None
        if row["Age"]:
            swimage = int(row["Age"])

        if "DQ" in timestr:
            return

        if "NT" in timestr:
            return

        if dist == "25m":
            return

        if relay == "Yes":
            return

        if pool not in ("50", "25"):
            return

        if gala:
            location = gala
        elif location:
            pass
        else:
            location = "Unknown"

        if dist not in DISTANCE:
            debug(f"Line {count}: Unknown distance {dist}", 1)
            return

        if stroke not in STROKES:
            debug(f"Line {count}: Unknown stroke {stroke}", 1)
            return

        verify = get_config(self.scm, C_RECORDS, C_VERIFY)
        age_eoy = get_config(self.scm, C_RECORDS, C_AGE_EOY)
        se_only = get_config(self.scm, C_RECORDS, C_SE_ONLY)
        all_ages = get_config(self.scm, C_RECORDS, C_ALL_AGES)

        member = None
        if asa not in self.scm.members.by_asa:
            debug(f"Line {count}: No SE Number {swimmer}", 2)
            # We can't check, so go with it...
            if se_only:
                return
            verify = False
            age_eoy = False
        else:
            member = self.scm.members.by_asa[asa]
            swimmer = member.knownas  # for consistency of spelling

        if swimage and swimage >= 25:
            age_eoy = True  # Masters are always EOY

        swimdate = datetime.datetime.strptime(xdate, SCM_CSV_DATE_FORMAT)

        if member and age_eoy:
            yob = member.dob.year
            swimyear = swimdate.year
            swimage = swimyear - yob

        if verify and member and member.date_joined and (swimdate < member.date_joined):
            debug(f"Line {count}: Ignored, not a member at time of swim", 2)
            return

        if swimage is None:
            return

        if swimage < 18:
            start_age = int(swimage / 2) * 2
            end_age = start_age + 1
        elif swimage in (18, 19):
            if all_ages:
                start_age = 19
            else:
                start_age = 18
            end_age = 24
        else:
            start_age = int(swimage / 5) * 5
            # round it
            if start_age == 20:
                if all_ages:
                    start_age = 19
                else:
                    start_age = 18
                end_age = 24
            else:
                end_age = start_age + 4

        agegroup = f"{start_age}-{end_age}"

        if all_ages:
            if swimage <= 18:
                agegroup = str(swimage)
            ALL_AGES[agegroup] += 1
        else:
            AGES[agegroup] += 1

        event = f"{gender} {agegroup} {dist} {stroke} {pool}"

        swim = {
            S_EVENT: event,
            S_ASA: asa,
            S_NAME: swimmer,
            S_TIMESTR: timestr,
            S_FTIME: convert_time(timestr),
            S_LOCATION: location,
            S_DATE: xdate,
        }

        self.records.check_swim(swim)

        return


class Record:
    """Manage a records file."""

    # pylint: disable=too-many-instance-attributes

    def __init__(self):
        """Initilaise Records handling."""

        self._filename = None
        self.scm = None
        self.records = {}
        self.swimmers = {}
        self.newrecords = {}
        self.fieldnames = None

        self.worldrecord = 0
        self.europeanrecord = 0
        self.britishrecord = 0

        self.date = None

    def check_swim(self, swim):
        """Check a swim time to see if it as a record."""
        check_event = swim[S_EVENT]
        if check_event in self.records:
            event = self.records[check_event]
            if swim[S_FTIME] >= event[S_FTIME]:
                return

        self.records[check_event] = swim

        sloc = swim[S_LOCATION]
        newrec = (
            f"New record: {check_event}, {swim[S_NAME]}, {swim[S_TIMESTR]}, {sloc}\n"
        )
        self.newrecords[check_event] = newrec

        if get_config(self.scm, C_RECORDS, C_OVERALL_FASTEST):
            split_event = swim[S_EVENT].split()
            split_event[1] = OVERALL
            o_event = " ".join(str(item) for item in split_event)
            if o_event in self.records:
                event = self.records[o_event]
                if swim[S_FTIME] >= event[S_FTIME]:
                    return

            o_swim = swim.copy()
            o_swim[S_EVENT] = o_event
            self.records[o_event] = o_swim

    def check_row(self, row, count):
        """Check a row from the records file."""
        event = row[S_EVENT]
        test = event.split()

        if GENDER.get(test[0], None) is None:
            notify(f"Line {count}: unknown gender '{test[0]}'\n")
            return

        if get_config(self.scm, C_RECORDS, C_ALL_AGES):
            ages = ALL_AGES
        else:
            ages = AGES

        if ages.get(test[1], None) is None:
            notify(f"Line {count}: unknown age '{test[1]}'\n")
            return

        if test[2] not in DISTANCE:
            notify(f"Line {count}: unknown distance '{test[2]}'\n")
            return

        if STROKES.get(test[3], None) is None:
            notify(f"Line {count}: unknown event '{test[3]}'\n")
            return

        if COURSE.get(test[4], None) is None:
            notify(f"Line {count}: unknown course '{test[4]}'\n")
            return

        if event in self.records:
            notify(f"Line {count}: duplicate '{event}'\n")
            return

        ages[test[1]] += 1
        row[S_FTIME] = convert_time(row[S_TIMESTR])
        self.records[event] = row

    def read_baseline(self, filename, scm):
        """Read Facebook file."""
        self._filename = filename
        self.scm = scm

        try:
            count = 0
            fdate = os.path.getmtime(filename)
            self.date = time.strftime(PRINT_DATE_FORMAT, time.localtime(fdate))

            with open(filename, newline="") as csvfile:
                csv_reader = csv.DictReader(csvfile, skipinitialspace=True)
                self.fieldnames = csv_reader.fieldnames

                for row in csv_reader:
                    count += 1
                    self.check_row(row, count)

            notify(f"Read {filename} ({count})...\n")
            return True

        except EnvironmentError as error:
            notify(f"Cannot open records file: {filename}\n{error}\n")
            return False

        except csv.Error as error:
            notify(f"Error in records file: {filename}\n{error}\n")
            return False

        notify("Unknown error\n")
        return False

    def write_records(self):
        """Write the new reords, and backup old."""

        try:
            home = str(Path.home())
            today = datetime.datetime.now()
            str_today = today.strftime("%y%m%d-%H%M%S")
            self.date = today.strftime(PRINT_DATE_FORMAT)

            filename = f"records-{str_today}.csv"

            mybackupdir = os.path.join(home, CONFIG_DIR, RECORDS_DIR, "backup")

            if os.path.exists(mybackupdir) is False:
                os.mkdir(mybackupdir)

            backupfile = os.path.join(home, mybackupdir, filename)
            copyfile(self._filename, backupfile)

        except EnvironmentError as error:
            notify(f"Error creating records backup file: {backupfile}\n{error}\n")
            return

        try:
            with open(self._filename, "w", newline="") as file:
                writer = csv.DictWriter(file, self.fieldnames, extrasaction="ignore")
                writer.writeheader()
                for record in sorted(self.records):
                    writer.writerow(self.records[record])

        except EnvironmentError as error:
            notify(f"Cannot open records file to update: {self._filename}\n{error}\n")
            return

        except csv.Error as error:
            notify(f"Error in writing records file: {self._filename}\n{error}\n")
            return

        notify(f"Updated records file: {self._filename}\n")

    def create_html(self, arg_gender, arg_strokes, arg_ages, arg_relay):
        """create a records file."""
        # Get prefix

        # pylint: disable=too-many-locals
        # pylint: disable=too-many-nested-blocks
        # pylint: disable=too-many-branches
        # pylint: disable=too-many-statements

        header = os.path.splitext(self._filename)[0]
        header += HEADER

        try:
            with open(header, FILE_READ) as file:
                prefix = file.read()

        except EnvironmentError as error:
            notify(f"records header file not found: {header}\n{error}\n")
            prefix = ""

        res = prefix

        overall = get_config(self.scm, C_RECORDS, C_OVERALL_FASTEST)
        if overall is None:
            overall = False

        for gender in arg_gender:
            o_gender = ""
            for stroke in arg_strokes:
                o_dist = ""

                if arg_relay:
                    loopdist = RELAY_DISTANCE
                else:
                    loopdist = STROKE_DISTANCE[stroke]

                for dist in loopdist:
                    o_age = ""
                    tag = TAG_INNER_ODD
                    for age in arg_ages:

                        if overall is False:
                            if age == OVERALL:
                                continue

                        if (arg_relay is False) and (arg_ages[age] == 0):
                            continue

                        opt = ""
                        opt_sc = self.print_record(stroke, dist, age, "25", gender)
                        opt_lc = self.print_record(stroke, dist, age, "50", gender)

                        if opt_sc or opt_lc:
                            if arg_relay:
                                opt += WRAP_TABLE_OPEN
                            if opt_sc:
                                opt += opt_sc
                            if opt_lc:
                                opt += opt_lc
                            if arg_relay:
                                opt += WRAP_TABLE_CLOSE
                        else:
                            opt += " -\n"

                        if tag == TAG_INNER_EVEN:
                            tag = TAG_INNER_ODD
                        else:
                            tag = TAG_INNER_EVEN

                        html = INNER_WRAP
                        html = re.sub("XX_TAG_XX", tag, html)
                        if arg_relay:
                            html = re.sub("XX_AGE_XX", RELAY_AGES[age], html)
                        else:
                            html = re.sub("XX_AGE_XX", age, html)
                        html = re.sub("XX_OPT_XX", opt, html)
                        o_age += html

                    if o_age:
                        html = AGE_WRAP
                        if arg_relay:
                            html = re.sub("XX_DIST_XX", PRINT_DISTANCE[dist], html)
                        else:
                            html = re.sub("XX_DIST_XX", dist, html)
                        html = re.sub("XX_AGE_XX", o_age, html)
                        o_dist += html

                if o_dist:
                    html = DIST_WRAP
                    html = re.sub("XX_DIST_XX", o_dist, html)
                    html = re.sub("XX_STROKE_XX", stroke.lower(), html)
                    o_gender += html

            if res:
                pgen = arg_gender[gender].lower()

                html = GENDER_WRAP
                html = re.sub("XX_GENDER_XX", pgen, html)
                html = re.sub("XX_O_GENDER_XX", o_gender, html)
                res += html

        res = self.print_extra(res)
        return res

    def print_extra(self, res):
        """Some extra printing..."""
        num = 0
        top10 = "<p>Top 10 record holders:</p><ul>"
        for holder in sorted(self.swimmers.items(), key=lambda x: x[1], reverse=True):
            top10 += f"<li>{holder[0]}: {holder[1]}</li>"
            num += 1
            if num > 10:
                break
        top10 += "</ul>"

        tally = "<ul>"
        if self.worldrecord > 0:
            tally += f"<li>World Records: {self.worldrecord}</li>\n"
        if self.europeanrecord > 0:
            tally += f"<li>European Records: {self.europeanrecord}</li>\n"
        if self.britishrecord > 0:
            tally += f"<li>British Records: {self.britishrecord}</li>\n"
        tally += "</ul>"

        res = res.replace("##RECORDHOLDERS##", top10)
        res = res.replace("##RECORDTALLY##", tally)
        res = res.replace("##RECORDUPDATE##", self.date)
        res += FOOTER
        return res

    def print_record(self, stroke, dist, age, course, gender):
        """Print a record."""

        # pylint: disable=too-many-arguments

        lookup = f"{gender} {age} {dist} {stroke} {course}"

        if lookup in self.records:
            record = self.records[lookup]
            xtime = record[S_TIMESTR]
            xdate = record[S_DATE]
            name = record[S_NAME]
            location = record[S_LOCATION]

            swimmer = name.split(",")
            if len(swimmer) == 2:
                name = f"{swimmer[1]} {swimmer[0]}"

            if name in self.swimmers:
                self.swimmers[name] += 1
            else:
                self.swimmers[name] = 1

            if re.search("world record", location, re.IGNORECASE):
                self.worldrecord += 1

            if re.search("european record", location, re.IGNORECASE):
                self.europeanrecord += 1

            if re.search("british record", location, re.IGNORECASE):
                self.britishrecord += 1

            location = re.sub(r"19\d\d", "", location)  # get rid of date
            location = re.sub(r"20\d\d", "", location)
            location = re.sub(r"\(25m\)", "", location)
            location = re.sub("25m", "", location)
            location = re.sub("50m", "", location)

            if xtime:
                opt = "\n  <div class=divTableRow>\n"
                opt += f"  <div class=divTableCell>{xtime} ({COURSE[course]})</div>\n"
                opt += f"  <div class=divTableCell>{xdate}</div>\n"
                opt += f"  <div class=divTableCell>{name}</div>\n"
                opt += f"  <div class=divTableCell>{location}</div>\n"
                opt += "  </div>\n"
                return opt
        return ""


class RelayRecord(Record):
    """Manage a relay records file."""

    def check_row(self, row, count):
        """Check a row from the records file."""

        event = row[S_EVENT].split()
        if RELAY_GENDER.get(event[0], None) is None:
            notify(f"Line {count}: unknown gender '{event[0]}'\n")
        if RELAY_AGES.get(event[1], None) is None:
            notify(f"Line {count}: unknown age '{event[1]}'\n")

        dist = RELAY_DISTANCE.get(event[2], None)
        if dist is None:
            mdist = RELAY_DISTANCE.get(event[2].rstrip("m"), None)
            if mdist is None:
                notify(f"Line {count}: unknown distance '{event[2]}'\n")
            else:
                event[2] = str(mdist)

        if RELAY_STROKES.get(event[3], None) is None:
            notify(f"Line {count}: unknown event '{event[3]}'\n")
        if COURSE.get(event[4], None) is None:
            notify(f"Line {count}: unknown course '{event[4]}'\n")

        jevent = " ".join(event)

        if jevent in self.records:
            notify(f"Line {count}: duplicate '{jevent}'\n")
            return

        self.records[jevent] = row

    def print_extra(self, res):
        """Some extra printing."""
        res = res.replace("##RECORDUPDATE##", self.date)
        res += FOOTER
        return res

    def print_record(self, stroke, dist, age, course, gender):
        """Print a record."""

        # pylint: disable=too-many-arguments
        # pylint: disable=too-many-locals

        lookup = f"{gender} {age} {dist} {stroke} {course}"
        if lookup in self.records:
            record = self.records[lookup]
            xtime = record[R_TIME]
            xdate = record[R_DATE]
            loc = record[R_LOCATION]
            swimmer1 = record[R_SWIMMER1]
            swimmer2 = record[R_SWIMMER2]
            swimmer3 = record[R_SWIMMER3]
            swimmer4 = record[R_SWIMMER4]
            swimmers = f"{swimmer1}, {swimmer2}, {swimmer3}, {swimmer4}"
            if xtime:
                opt = "\n <div class=divTableRow>\n"
                opt += f" <div class=divTableCell>{xtime} ({COURSE[course]})</div>\n"
                opt += f" <div class=divTableCell>{xdate}</div>\n"
                opt += " <div class=divTable><div class=divTableBody>"
                opt += "<div class=divTableRow>\n"
                opt += f"  <div class=divTableCell>{loc}</div>\n"
                opt += " </div><div class=divTableRow>\n"
                opt += f" <div class=divTableCell>{swimmers}</div>\n"
                opt += " </div></div></div></div>\n"
                return opt
        return ""


def convert_time(xtime):
    """Convert a time to a number of seconds."""

    try:
        hms = xtime.split(":")
        if len(hms) == 2:
            res = float(hms[0]) * 60 + float(hms[1])
        else:
            res = float(hms[0])
        return res
    except ValueError:
        debug(f"invalid time {xtime} ", 3)
        return 999999


# pylint: disable=too-many-lines


DEFAULT_RECORDS = """event,name,time,location,date
"""

DEFAULT_HEADER = """<style>

input[type=radio],
input[type=checkbox] {
  display: none;
}

input[type=radio]+label,
input[type=checkbox]+label {
  display: inline-block;
  margin: -2px;
  padding: 4px 1em;
  margin-bottom: 0;
  font-size: 12px;
  line-height: 19px;
  color: #333;
  text-align: center;
  vertical-align: middle;
  cursor: pointer;
  background-color: #f5f5f5;
  background-image: -moz-linear-gradient(top, #fff, #e6e6e6);
  background-image: -webkit-gradient(linear, 0 0, 0 100%, from(#fff), to(#e6e6e6));
  background-image: -webkit-linear-gradient(top, #fff, #e6e6e6);
  background-image: -o-linear-gradient(top, #fff, #e6e6e6);
  background-image: linear-gradient(to bottom, #fff, #e6e6e6);
  background-repeat: repeat-x;
  border: 1px solid #ccc;
  border-color: #e6e6e6 #e6e6e6 #bfbfbf;
  border-color: rgba(0, 0, 0, 0.1) rgba(0, 0, 0, 0.1) rgba(0, 0, 0, 0.25);
  border-bottom-color: #b3b3b3;
  filter: progid: DXImageTransform.Microsoft.gradient(startColorstr='#ffffffff',
            endColorstr='#ffe6e6e6', GradientType=0);
  filter: progid: DXImageTransform.Microsoft.gradient(enabled=false);
  -webkit-box-shadow: inset 0 1px 0 rgba(255, 255, 255, 0.2),
           0 1px 2px rgba(0, 0, 0, 0.05);
  -moz-box-shadow: inset 0 1px 0 rgba(255, 255, 255, 0.2),
           0 1px 2px rgba(0, 0, 0, 0.05);
  box-shadow: inset 0 1px 0 rgba(255, 255, 255, 0.2),
           0 1px 2px rgba(0, 0, 0, 0.05);
}

input[type=radio]:checked+label,
input[type=checkbox]:checked+label {
  background-image: none;
  outline: 0;
  -webkit-box-shadow: inset 0 2px 4px rgba(0, 0, 0, 0.15),
              0 1px 2px rgba(0, 0, 0, 0.05);
  -moz-box-shadow: inset 0 2px 4px rgba(0, 0, 0, 0.15),
              0 1px 2px rgba(0, 0, 0, 0.05);
  box-shadow: inset 0 2px 4px rgba(0, 0, 0, 0.15), 0 1px 2px rgba(0, 0, 0, 0.05);
  background-color: #4CAF50;
  color: white;
}


.divTable{
  display: table;
}
.divTableRow {
  display: table-row;
}

.divTableCell {
  display: table-cell;
  border: 0px;
  font-size: 12px;
  display: table-cell;
  padding: 3px 10px;
  vertical-align: middle
}

[name=record-inner-even] .divTableCell {
  display: table-cell;
  background-color: #83c983;
  color: black;
}
[name=record-inner-odd] .divTableCell {
  display: table-cell;
  background-color: #b8e0b8;
  color: black;
}

.divTableBody {
  display: table-row-group;
}

</style>


<script>

function SetBlock(block, state) {

    var elms = document.querySelectorAll(block);
    for (var i = 0; i < elms.length; i++) {
        elms[i].style.display = state;
    }
}

function ClearRecord() {

    SetBlock('[name="record-medley"]', "none")
    SetBlock('[name="record-free"]', "none")
    SetBlock('[name="record-fly"]', "none")
    SetBlock('[name="record-breast"]', "none")
    SetBlock('[name="record-back"]', "none")
    SetBlock('[name="record-male"]', "none")
    SetBlock('[name="record-female"]', "none")

}

function ShowBlock(block) {

    var str = document.querySelector(block).value;

    var target = '[name=\"' + str + '\"]'
    var elms = document.querySelectorAll(target);
    for (var i = 0; i < elms.length; i++) {

        if (elms[i].style.display === "none") {
            elms[i].style.display = "block";
        } else {
            elms[i].style.display = "none";
        }
    }
}


function ShowRecord() {

    ClearRecord();

    ShowBlock('input[name="record-stroke"]:checked')
    ShowBlock('input[name="record-gender"]:checked')

}


function docReady(fn) {

    // see if DOM is already available
    if (document.readyState === "complete" || document.readyState === "interactive") {
        // call on next available tick
        setTimeout(fn, 1);
    } else {
        document.addEventListener("DOMContentLoaded", fn);
    }
}

function SetButton(btn) {
    var h = document.getElementById(btn);
    h.onchange = function() {
        ShowRecord()
    };
}


docReady(function() {

    ClearRecord();

    SetBlock('[name="record-free"]', 'block')
    SetBlock('[name="record-male"]', 'block')

    SetButton('btn-free')
    SetButton('btn-back')
    SetButton('btn-breast')
    SetButton('btn-fly')
    SetButton('btn-medley')
    SetButton('btn-male')
    SetButton('btn-female')


});


</script>

<h3>Swimming Records</h3>

<div class="row clearfix">
<div class="column two-third">

<p>Rules: Must have been a Club member at the time of the swim.
Age as of December 31st in the year of the swim.</p>

##RECORDTALLY##
</div>
<div class="column third">
##RECORDHOLDERS##
</div>
</div>


<h2>Records</h2>
<p>Last Update: ##RECORDUPDATE##</p>
<p>Event:
  <input checked="checked" class="radio-group" id="btn-free"
  name="record-stroke" type="radio" value="record-free" />
  <label for="btn-free">Freestyle</label>

   <input class="radio-group" id="btn-fly" name="record-stroke"
   type="radio" value="record-fly" />
  <label for="btn-fly">Fly</label>

   <input class="radio-group" id="btn-back" name="record-stroke"
   type="radio" value="record-back" />
  <label for="btn-back">Back</label>

   <input class="radio-group" id="btn-breast" name="record-stroke"
   type="radio" value="record-breast" />
  <label for="btn-breast">Breast</label>

  <input class="radio-group" id="btn-medley" name="record-stroke"
  type="radio" value="record-medley" />
  <label for="btn-medley">Individual Medley</label>
</p>

<p>Gender:
  <input checked="checked" class="radio-group" id="btn-male" name="record-gender"
  type="radio" value="record-male" />
  <label for="btn-male">Male</label>

  <input class="radio-group" id="btn-female" name="record-gender"
  type="radio" value="record-female" />
  <label for="btn-female">Female</label>
</p>


<!-- Text below script generated. -->

"""



"""Installation of scm-helper."""
import setuptools

with open("README.md", "r") as fh:
    DESCRIPTION = fh.read()

VERSION = {}
with open("scm_helper/version.py") as fh:
    exec(fh.read(), VERSION)

setuptools.setup(
    name="scm-helper",
    version=VERSION["VERSION"],
    author="Colin Robbins",
    author_email="colin.john.robbins@gmail.com",
    description="Helper tool to manage data in Swim Club Manager",
    long_description=DESCRIPTION,
    long_description_content_type="text/markdown",
    url="https://github.com/ColinRobbins/scm-helper",
    entry_points={"console_scripts": ["scm=scm_helper.main:main"]},
    project_urls={
        "Bug Tracker": "https://github.com/ColinRobbins/scm-helper/issues",
        "Documentation": "https://github.com/ColinRobbins/scm-helper/wiki",
        "Source Code": "https://github.com/ColinRobbins/scm-helper",
        "Change Log": "https://github.com/ColinRobbins/scm-helper/blob/master/CHANGELOG.md",
    },
    packages=setuptools.find_packages(),
    install_requires=[
        "schema>=0.7.2",
        "requests>=2.22.0",
        "pyyaml>=5.3",
        "cryptography>=2.8",
        "func-timeout>=4.3.5",
        "selenium>=3.141.0",
    ],
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    python_requires=">=3.7",
)



"""SCM Conduct."""
from scm_helper.config import (
    A_DATEAGREED,
    A_GUID,
    A_MEMBERS,
    C_CONDUCT,
    C_IGNORE_GROUP,
    C_LISTS,
    C_TYPES,
    CTYPE_COACH,
    CTYPE_COMMITTEE,
    CTYPE_PARENT,
    CTYPE_POLO,
    CTYPE_SWIMMER,
    CTYPE_SYNCHRO,
    CTYPE_VOLUNTEER,
    get_config,
)
from scm_helper.entity import Entities, Entity
from scm_helper.issue import E_NO_CONDUCT, E_NO_CONDUCT_DATE, debug_trace, issue
from scm_helper.notify import notify


class CodesOfConduct(Entities):
    """Conduct."""

    def get_data(self):
        """
        Get member data.

        For some reason the API behaves differently,
        we need to read each one...
        """
        self._raw_data = []

        notify(f"{self._name}... ")

        entities = self.scm.api_read(self._url, 1)
        if entities is None:
            return False
        page = 1

        for entity in entities:
            guid = entity["Guid"]
            notify(f"{page} ")

            api_data = self.scm.api_read(f"{self._url}/{guid}", 1)
            if api_data is None:
                return False
            self._raw_data += [api_data]

            data = self.new_entity(api_data)
            self.entities.append(data)
            self.by_guid[data.guid] = data
            self.by_name[data.name] = data

            if data.is_active:
                self.count += 1
            page += 1

        notify("\n")

        return True

    def new_entity(self, entity):
        """Create a new entity."""
        return Conduct(entity, self.scm, self._url)

    @debug_trace(5)
    def analyse(self):
        """Analyse the conduct class."""
        if get_config(self.scm, C_CONDUCT) is None:
            return

        for code in self.entities:
            code.analyse()


class Conduct(Entity):
    """A conduct instance."""

    def __init__(self, entity, scm, url):
        """Initialise."""
        super().__init__(entity, scm, url)
        self._raw_data = entity

    def linkage(self, members):
        """Link code."""
        super().linkage(members)
        for member in self.members:
            member.add_conduct(self)

    def analyse(self):
        """Analyse the conduct entry."""
        # A better way of doing this would be to add
        # the attribute to the swimmer in linkage.
        # This approach breaks the model. Oh well.

        ignores = get_config(self.scm, C_CONDUCT, self.name, C_IGNORE_GROUP)

        for member in self.data[A_MEMBERS]:

            if member[A_DATEAGREED]:
                continue

            person = self.scm.members.by_guid[member[A_GUID]]
            if person.confirmed_date:  # Will get a not confirmed error later in not set

                found_ignore = False
                if ignores:
                    for ignore in ignores:
                        if person.find_group(ignore):
                            found_ignore = True
                if found_ignore:
                    continue

                issue(person, E_NO_CONDUCT_DATE, self.name, 0, person.first_group)
                codes = get_config(self.scm, C_LISTS, C_CONDUCT)
                if codes:
                    for code in codes:
                        if self.name == code:
                            msg = f"{self.name} missing"
                            self.scm.lists.add(msg, person)

    @property
    def name(self):
        """Name."""
        return self.data["Title"]


# Outside of class
def check_conduct(member, my_codes):
    """Analyse a code of conduct."""
    # pylint: disable=too-many-branches

    if get_config(member.scm, C_CONDUCT) is None:
        return

    type_dict = {
        CTYPE_SWIMMER: member.is_swimmer,
        CTYPE_PARENT: member.is_parent,
        CTYPE_COACH: member.is_coach,
        CTYPE_COMMITTEE: member.is_committee_member,
        CTYPE_VOLUNTEER: member.is_volunteer,
        CTYPE_POLO: member.is_polo,
        CTYPE_SYNCHRO: member.is_synchro,
    }

    codes = member.scm.conduct.entities

    for code in codes:
        ignores = get_config(member.scm, C_CONDUCT, code.name, C_IGNORE_GROUP)

        found_ignore = False
        if ignores:
            for ignore in ignores:
                if member.find_group(ignore):
                    found_ignore = True
        if found_ignore:
            continue

        types = get_config(member.scm, C_CONDUCT, code.name, C_TYPES)
        if types is None:
            return

        for atype in types:
            found = False
            func = type_dict[atype]
            if func is True:
                for my_code in my_codes:
                    if my_code == code:
                        found = True
                        break
                if found is False:
                    issue(member, E_NO_CONDUCT, f"{code.name}")

                    # TODO - Need SCM bug fix first
                    # if code.newdata and (A_MEMBERS in code.newdata):
                    #    fix = code.newdata
                    # else:
                    #    fix = {}
                    #    fix[A_MEMBERS] = code.data[A_MEMBERS].copy()
                    # add = {A_GUID: member.guid}
                    # fix[A_MEMBERS].append(add)
                    # code.fixit(fix, f"Add {member.name}")



"""Read and process a file of Facebook date."""
import ntpath
import os
import re
from pathlib import Path

from scm_helper.config import (
    C_FACEBOOK,
    C_FILES,
    C_GROUPS,
    CONFIG_DIR,
    FILE_READ,
    get_config,
)
from scm_helper.files import Files
from scm_helper.notify import notify

# should probably parse the structure, but a simple regex seems to work...
FACEBOOK_RE = re.compile(
    'ref=MEMBER_LIST" rel="dialog" title="([a-zA-Z -]*)" tabindex='
)


class Facebook:
    """Read and process all facebook files."""

    def __init__(self):
        """Initialise."""
        self.scm = None
        self.facebook = []

    def read_data(self, scm):
        """Read each file."""
        self.scm = scm

        home = str(Path.home())
        mydir = os.path.join(home, CONFIG_DIR)

        cfg = get_config(scm, C_FACEBOOK, C_FILES)
        if cfg:
            for facebook in cfg:
                face = FacebookPage()
                filename = os.path.join(mydir, facebook)
                res = face.readfile(filename, scm)
                if res:
                    self.facebook.append(face)
                    if face.parse() is False:
                        return False
                else:
                    return False
            return True

        cfg = get_config(scm, C_FACEBOOK, C_GROUPS)
        if cfg:
            for facebook in cfg:
                face = FacebookPage()
                res = face.readurl(facebook, scm)
                if res:
                    self.facebook.append(face)
                else:
                    return False

            return True

        return False

    def analyse(self):
        """Analyse filebook files."""
        for facebook in self.facebook:
            facebook.analyse()

    def print_errors(self):
        """Print errors."""
        res = ""
        for facebook in self.facebook:
            res += facebook.print_errors()

        if res == "":
            res = "Nothing to report"
        return res

    def delete(self):
        """Delete."""
        for facebook in self.facebook:
            del facebook


class FacebookPage(Files):
    """Manage a facebook file."""

    def __init__(self):
        """Initilaise Facebook."""
        super().__init__()
        self._filename = None
        self.scm = None
        self.data = None
        self.users = []

    def readfile(self, filename, scm):
        """Read Facebook file."""
        self._filename = filename
        self.scm = scm

        shortfile = ntpath.basename(filename)

        notify(f"Reading {shortfile}...\n")

        try:
            with open(filename, FILE_READ, encoding="utf8") as file:
                self.data = file.read().replace("\n", "")
            file.close()
            return True

        except EnvironmentError:
            notify(f"Cannot open facebook file: {filename}\n")
            return False

    def readurl(self, url, scm):
        """Read Facebook page."""
        self._filename = url.rstrip("/")  # Remove '/' for result printing

        self.scm = scm

        if self.scm.ipad:
            notify("Not implemented on iPad")
            return False

        # pylint: disable=import-outside-toplevel
        from scm_helper.browser import fb_read_url

        res = fb_read_url(scm, url)

        if res is None:
            return False

        self.users = res
        return True

    def parse(self):
        """Parse file to find strings."""
        users = FACEBOOK_RE.findall(self.data)

        if users is None:
            notify("Could not parse Facebook file")
            return False

        count = 0
        for user in users:
            count += 1
            self.users.append(user)

        notify(f"Read {count} members.\n")

        if count == 0:
            notify("Error: No members found")
            return False
        return True

    def analyse(self):
        """Analyse Facebook file."""
        members = self.scm.members.by_name
        knownas = self.scm.members.knownas
        facebook = self.scm.members.facebook

        for user in self.users:
            inactive = False
            if user in members:
                if members[user].is_active:
                    continue
                inactive = True

            if user in knownas:
                if knownas[user].is_active:
                    continue
                inactive = True

            if user in facebook:  # NB this can override inactive = True above
                # could be active in a different entry (eg parent)
                if facebook[user].is_active:
                    continue
                inactive = True

            if inactive:
                self.file_error(user, "Facebook user is inactive in SCM (resigned?)")
            else:
                self.file_error(user, "Facebook user not in SCM")



"""SCM Members."""
from scm_helper.config import (
    A_ACTIVE,
    A_FIRSTNAME,
    A_LASTNAME,
    C_NAME,
    C_TYPES,
    CTYPE_SYNCHRO,
    get_config,
)
from scm_helper.entity import Entities
from scm_helper.issue import E_DUPLICATE, debug, issue
from scm_helper.member import Member
from scm_helper.notify import notify


class Members(Entities):
    """Members."""

    # pylint: disable=too-many-instance-attributes
    # Need them all!

    def __init__(self, scm, name, url):
        """Initialize."""
        # pylint: disable=W0231
        # DO NOT call super() - it re-reads the data in the wrong class.
        # pylint: disable=too-many-instance-attributes
        # Need them all!
        self.entities = []
        self.by_guid = {}
        self.by_name = {}
        self.knownas = {}
        self.by_asa = {}
        self._name = name
        self._url = url
        self._raw_data = []

        self.facebook = {}
        self.count_coaches = 0
        self.count_parents = 0
        self.count_inactive = 0
        self.count_swimmers = 0
        self.count_waterpolo = 0
        self.count_synchro = 0
        self.count_volunteer = 0
        self.count_not_confirmed = 0
        self.count = 0

        self.scm = scm

    def check_duplicate(self, member):
        """See if member already exists before adding."""
        firtname = member[A_FIRSTNAME]
        lastname = member[A_LASTNAME]
        name = f"{firtname} {lastname}"
        if name in self.by_name:
            if member[A_ACTIVE] == "1" and self.by_name[name].is_active:
                act1 = member[A_ACTIVE]
                act2 = self.by_name[name].is_active
                debug(f"{name}: {act1}-{act2}", 6)
                issue(self.by_name[name], E_DUPLICATE, name)
            else:
                active = self.by_name[name].is_active
                if member[A_ACTIVE] == "0" and active is False:
                    issue(self.by_name[name], E_DUPLICATE, "BOTH inactive", 9)
                else:
                    issue(self.by_name[name], E_DUPLICATE, "One is inactive", -1)
            return
        if name in self.knownas:
            if member[A_ACTIVE] == "1" and self.knownas[name].is_active:
                issue(self.knownas[name], E_DUPLICATE, name, 0, "(Known as)")
            else:
                issue(self.knownas[name], E_DUPLICATE, "One is inactive (Known as)", -1)

    def create_entities(self, entities):
        """Create a member objects."""
        i = 0
        for member in entities:
            self.check_duplicate(member)

            data = Member(member, self.scm, self._url)
            self.entities.append(data)
            self.by_guid[data.guid] = data
            self.by_name[data.name] = data
            self.knownas[data.knownas] = data
            if data.asa_number:
                self.by_asa[data.asa_number] = data
            if data.facebook:
                for face in data.facebook:
                    self.facebook[face] = data
            if data.is_active:
                if data.is_coach:
                    self.count_coaches += 1
                if data.is_parent:
                    self.count_parents += 1
                if data.is_swimmer:
                    self.count_swimmers += 1
                if data.is_polo:
                    self.count_waterpolo += 1
                if data.is_synchro:
                    self.count_synchro += 1
                if data.is_volunteer:
                    self.count_volunteer += 1
                self.count += 1
            else:
                self.count_inactive += 1
            i += 1

        return i

    def find_member(self, find):
        """Find a member."""
        if find in self.by_name:
            return self.by_name[find]
        if find in self.knownas:
            return self.knownas[find]
        return None

    def linkage(self):
        """Create Member links."""
        for member in self.entities:
            member.linkage(self.scm.members)

        for member in self.entities:
            # pylint: disable=fixme
            # Fix API error. not all parent links are returned ,so reverse link
            # TODO remove when API fixed.
            member.linkage2()

    def fix_search(self):
        """fix_search_index."""
        for member in self.entities:
            res = member.fix_search()
            if res is False:
                return False
        return True

    def print_notes(self):
        """Print the notes."""
        res = ""
        for member in self.entities:
            res += member.print_notes()
        return res

    def se_check(self):
        """Check against an SE online."""
        if self.scm.ipad:
            notify("Not implemented on iPad")
            return False

        # pylint: disable=import-outside-toplevel
        from scm_helper.browser import se_check

        return se_check(self.scm, self.entities)

    def print_summary(self):
        """Print a summary."""
        name = get_config(self.scm, C_TYPES, CTYPE_SYNCHRO, C_NAME)

        opt = ""

        opt += f"Members: {self.count}\n"
        opt += f"   Swimmers: {self.count_swimmers}\n"
        opt += f"   {name}: {self.count_synchro}\n"
        opt += f"   Water Polo: {self.count_waterpolo}\n"
        opt += f"   Volunteers: {self.count_volunteer}\n"
        opt += f"   Coaches: {self.count_coaches}\n"
        opt += f"   Parents: {self.count_parents}\n"
        opt += f"   Inactive: {self.count_inactive}\n"

        return opt

    def print_swimmers_sessions(self):
        """Print sessions each swimmer is in."""
        res = "Name,Group,"
        for session in self.scm.sessions.entities:
            if session.is_active:
                res += f'"{session.full_name}",'
        res += "\n"

        for swimmer in self.entities:
            if swimmer.is_active:
                sessions = swimmer.print_swimmer_sessions(True)
                if sessions:
                    res += f"{swimmer.name},{swimmer.first_group},"
                    res += sessions
                    res += "\n"

        return res

    @property
    def url(self):
        """Return URL."""
        return self._url



"""Windows GUI."""
# pylint: disable=too-many-lines
import os.path
import sys
import threading
import traceback
import webbrowser
from datetime import datetime
from pathlib import Path
from tkinter import (
    DISABLED,
    END,
    NORMAL,
    Button,
    E,
    Entry,
    Frame,
    Label,
    LabelFrame,
    Menu,
    N,
    OptionMenu,
    S,
    StringVar,
    Toplevel,
    W,
    filedialog,
    messagebox,
    scrolledtext,
)

from func_timeout import FunctionTimedOut, func_timeout

from scm_helper.api import API
from scm_helper.config import (
    BACKUP_DIR,
    CONFIG_DIR,
    CONFIG_FILE,
    FILE_READ,
    FILE_WRITE,
    HELPURL,
    check_default,
)
from scm_helper.facebook import Facebook
from scm_helper.file import Csv
from scm_helper.issue import REPORTS, IssueHandler, debug
from scm_helper.license import LICENSE
from scm_helper.notify import set_notify
from scm_helper.records import Records
from scm_helper.version import VERSION

NSEW = N + S + E + W
AFTER = 2500  # How log to wait for Windows to catchup after processing


class ScmGui:
    """Tkinter based GUI (for Windows)."""

    # pylint: disable=too-many-instance-attributes
    # pylint: disable=too-many-public-methods

    def __init__(self, master):
        """Initialise."""
        # pylint: disable=too-many-statements
        # super().__init__(None)
        self.master = master
        self.issue_window = None
        self.issue_text = None
        self.report_window = None
        self.report_text = None
        self.issues = None
        self.scm = None
        self.gotdata = False
        self.thread = None
        self.grouping = None
        self.reports = None
        self.menus = []
        self.menu_fixit = None
        self.notify_text = None

        master.title("SCM Helper")

        self.create_main_window()
        self.create_menu()

        set_notify(self.notify_text)
        self.issues = IssueHandler()
        self.scm = API(self.issues)

        if self.scm.get_config_file() is False:
            msg = "Error in config file - see status window for details."
            messagebox.showerror("Error", msg, parent=self.master)
            cfgmsg = ""
        else:
            cfgmsg = check_default(self.scm)

        self.api_init = False

        msg = "Welcome to SCM Helper by Colin Robbins.\n"
        msg += cfgmsg
        msg += "Please enter your password.\n"
        self.notify_text.insert(END, msg)
        self.notify_text.config(state=DISABLED)

    def create_main_window(self):
        """Create the main window."""
        top_frame = Frame(self.master)
        top_frame.grid(row=0, column=0, sticky=W + E)

        label = Label(top_frame, text="Password: ")
        label.grid(row=0, column=0, pady=10, padx=10)

        self.__password = StringVar()
        self.__password.set("")
        password = Entry(top_frame, show="*", textvariable=self.__password, width=20)
        password.grid(row=0, column=1, pady=10, padx=10)

        msg = "Analyse"
        self.button_analyse = Button(top_frame, text=msg, command=self.analyse_window)
        self.button_analyse.grid(row=0, column=2, pady=10, padx=10)

        self.master.bind("<Return>", self.analyse_enter)

        self.button_backup = Button(top_frame, text="Backup", command=self.backup)
        self.button_backup.grid(row=0, column=3, pady=10, padx=10)

        self.button_fixit = Button(top_frame, text="Fixit", command=self.fixit)
        self.button_fixit.config(state=DISABLED)
        self.button_fixit.grid(row=0, column=4, pady=10, padx=10)

        top_group = LabelFrame(self.master, text="Notifications...", pady=5, padx=5)
        top_group.grid(row=1, column=0, columnspan=5, pady=10, padx=10, sticky=NSEW)

        self.master.columnconfigure(0, weight=1)
        self.master.rowconfigure(1, weight=1)

        top_group.columnconfigure(0, weight=1)
        top_group.rowconfigure(0, weight=1)

        self.notify_text = scrolledtext.ScrolledText(top_group, width=60, height=20)
        self.notify_text.grid(row=0, column=0, sticky=NSEW)

    def create_menu(self):
        """Create Menus."""
        # pylint: disable=too-many-statements
        menubar = Menu(self.master)
        file = Menu(menubar, tearoff=0)
        file.add_command(label="Open Archive", command=self.open_archive)
        self.menus.append([file, "Open Archive"])
        file.add_separator()
        file.add_command(label="Exit", command=self.master.quit)
        menubar.add_cascade(label="File", menu=file)

        cmd = Menu(menubar, tearoff=0)
        cmd.add_command(label="Edit Config", command=self.edit_config)
        self.menus.append([cmd, "Edit Config"])
        cmd.add_separator()
        cmd.add_command(label="Create Lists", command=self.create_lists, state=DISABLED)
        self.menus.append([cmd, "Create Lists"])

        cmd.add_command(label="Fix Errors", command=self.fixit, state=DISABLED)
        self.menu_fixit = [cmd, "Fix Errors"]

        label = "Fix Search index"
        cmd.add_command(label=label, command=self.fix_search, state=DISABLED)
        self.menus.append([cmd, label])

        menubar.add_cascade(label="Edit", menu=cmd)

        cmd = Menu(menubar, tearoff=0)

        cmd.add_command(label="Analyse Facebook", command=self.facebook, state=DISABLED)
        self.menus.append([cmd, "Analyse Facebook"])

        label = "Analyse Swim England File"
        cmd.add_command(label=label, command=self.swim_england, state=DISABLED)
        self.menus.append([cmd, label])

        label = "Analyse Swim England Registrations Online"
        cmd.add_command(label=label, command=self.swim_england_online, state=DISABLED)
        self.menus.append([cmd, label])

        cmd.add_command(label="List Coaches", command=self.coaches, state=DISABLED)
        self.menus.append([cmd, "List Coaches"])

        label = "Show Swimmers Covid Declaration per session"
        cmd.add_command(label=label, command=self.covid, state=DISABLED)
        self.menus.append([cmd, label])

        label = "List Swimmers sessions"
        cmd.add_command(label=label, command=self.sessions, state=DISABLED)
        self.menus.append([cmd, label])

        label = "Show Not-confirmed Emails"
        cmd.add_command(label=label, command=self.confirm, state=DISABLED)
        self.menus.append([cmd, label])

        menubar.add_cascade(label="Reports", menu=cmd)

        record = Menu(menubar, tearoff=0)

        record.add_command(label="Process Records", command=self.process_records)

        label = "Read New Swim Times"
        record.add_command(label=label, command=self.new_times, state=DISABLED)
        self.menus.append([record, label])

        menubar.add_cascade(label="Records", menu=record)

        about = Menu(menubar, tearoff=0)
        about.add_command(label="About...", command=xabout)
        about.add_command(label="Help", command=xhelp)
        about.add_command(label="License", command=xlicense)
        menubar.add_cascade(label="About", menu=about)

        self.master.config(menu=menubar)

    def scm_init(self):
        """Initialise SCM."""
        self.notify_text.config(state=NORMAL)
        password = self.__password.get()

        self.clear_data()
        self.api_init = False

        if password:
            if self.scm.initialise(password) is False:
                messagebox.showerror("Error", "Cannot initialise SCM (wrong password?)")
                self.clear_data()
                self.notify_text.config(state=DISABLED)
                return False
            self.api_init = True
            return True

        messagebox.showerror("Password", "Please give a password!")
        self.notify_text.config(state=DISABLED)

        return False

    def prep_report(self, create=True, delete=True):
        """Prepare for a report."""
        if self.gotdata is False:
            messagebox.showerror("Error", "Run analyse first to collect data")
            return False

        if create and (self.report_window is None):
            self.create_report_window()

        if self.report_window:
            self.report_text.config(state=NORMAL)
            self.report_text.delete("1.0", END)

        self.notify_text.config(state=NORMAL)
        if delete:
            self.notify_text.delete("1.0", END)

        return True

    def open_archive(self):
        """Open Archive file."""
        if self.thread:
            return  # already running

        if self.scm_init() is False:
            return

        self.thread = AnalysisThread(self, True).start()

    def swim_england(self):
        """Process Swim England File."""
        if self.prep_report() is False:
            return

        csv = Csv()

        home = str(Path.home())
        cfg = os.path.join(home, CONFIG_DIR)

        dir_opt = {}
        dir_opt["title"] = "Find Swim England File"
        dir_opt["initialdir"] = cfg
        dir_opt["parent"] = self.report_window
        dir_opt["defaultextension"] = ".csv"

        where = filedialog.askopenfilename(**dir_opt)
        if csv.readfile(where, self.scm) is False:
            messagebox.showerror("Error", "Could not read CSV file")
            self.notify_text.config(state=DISABLED)
            self.report_text.config(state=DISABLED)
            return

        if wrap(10, csv.analyse, self.scm) is False:
            self.notify_text.config(state=DISABLED)
            self.report_text.config(state=DISABLED)
            return

        output = wrap(None, csv.print_errors)

        del csv
        self.report_text.insert(END, output)
        self.report_text.config(state=DISABLED)
        self.notify_text.config(state=DISABLED)
        self.report_window.lift()

    def swim_england_online(self):
        """Window for reports."""
        if self.thread:
            return  # already running

        if self.prep_report(False) is False:
            return

        self.thread = SwimEnglandThread(self).start()

    def facebook(self):
        """Window for reports."""
        if self.thread:
            return  # already running

        if self.prep_report(False) is False:
            return

        self.thread = FacebookThread(self).start()

    def confirm(self):
        """Confirm email Report."""
        if self.prep_report() is False:
            return

        output = self.issues.confirm_email()
        self.report_text.insert(END, output)
        self.report_text.config(state=DISABLED)
        self.notify_text.config(state=DISABLED)
        self.report_window.lift()

    def coaches(self):
        """Coaches Report."""
        if self.prep_report() is False:
            return

        output = wrap(None, self.scm.sessions.print_coaches)
        self.report_text.insert(END, output)
        self.report_text.config(state=DISABLED)
        self.notify_text.config(state=DISABLED)
        self.report_window.lift()

    def covid(self):
        """Covid Report."""
        if self.prep_report() is False:
            return

        output = wrap(None, self.scm.sessions.print_swimmers_covid)
        self.report_text.insert(END, output)
        self.report_text.config(state=DISABLED)
        self.notify_text.config(state=DISABLED)
        self.report_window.lift()

    def sessions(self):
        """Swimmers per Sessions Report."""
        if self.prep_report() is False:
            return

        output = wrap(None, self.scm.members.print_swimmers_sessions)
        self.report_text.insert(END, output)
        self.report_text.config(state=DISABLED)
        self.notify_text.config(state=DISABLED)
        self.report_window.lift()

    def edit_config(self):
        """Edit Config."""
        home = str(Path.home())
        cfg = os.path.join(home, CONFIG_DIR, CONFIG_FILE)

        Edit(self.master, cfg, self.scm, self)

    def create_lists(self):
        """Create Lists."""
        if self.gotdata is False:
            messagebox.showerror("Error", "Run analyse first to collect data")
            return

        if self.thread:
            return  # already running

        self.thread = UpdateThread(self).start()

    def analyse_window(self):
        """Window for analysis result."""
        if self.thread:
            return  # already running

        if self.scm_init() is False:
            return

        self.thread = AnalysisThread(self, False).start()

    def analyse_enter(self, event):
        """Window for analysis result."""
        debug(f"Event: {event}", 7)
        self.analyse_window()

    def fix_search(self):
        """Window for reports."""
        if self.thread:
            return  # already running

        if self.gotdata is False:
            messagebox.showerror("Error", "Analyse data first, before fixing (2)")
            return

        self.thread = SearchThread(self).start()

    def fixit(self):
        """Window for reports."""
        if self.thread:
            return  # already running

        if self.gotdata is False:
            messagebox.showerror("Error", "Analyse data first, before fixing")
            return

        self.set_buttons(DISABLED)

        length = len(self.scm.fixable)
        if length == 0:
            messagebox.showerror("Error", "Nothing to fix")
            self.set_buttons(NORMAL)
            return

        wrap(None, self.scm.apply_fixes)

        self.set_buttons(NORMAL)

    def backup(self):
        """Window for reports."""
        if self.thread:
            return  # already running

        if self.scm_init() is False:
            return

        self.thread = BackupThread(self).start()

    def clear_data(self):
        """Prepare to rerun."""
        wrap(None, self.scm.delete)
        self.gotdata = False

    def create_report_window(self):
        """Create the reports window."""
        self.report_window = Toplevel(self.master)
        self.report_window.title("SCM Helper - Report Window")

        top_frame = Frame(self.report_window)
        top_frame.grid(row=0, column=0, sticky=W + E)

        txt = "Report..."
        top_group = LabelFrame(self.report_window, text=txt, pady=5, padx=5)
        top_group.grid(row=0, column=0, pady=10, padx=10, sticky=NSEW)

        self.report_window.columnconfigure(0, weight=1)
        self.report_window.rowconfigure(0, weight=1)
        top_group.columnconfigure(0, weight=1)
        top_group.rowconfigure(0, weight=1)

        self.report_text = scrolledtext.ScrolledText(top_group, width=80, height=40)
        self.report_text.grid(row=0, column=0, sticky=NSEW)

        self.report_window.protocol("WM_DELETE_WINDOW", self.close_report_window)

    def close_report_window(self):
        """Close report window."""
        self.report_window.destroy()
        self.report_window = None

    def create_issue_window(self):
        """Create the results window."""
        self.issue_window = Toplevel(self.master)
        self.issue_window.title("SCM Helper - Issue Window")

        self.reports = StringVar()
        self.reports.set("All Reports")

        top_frame = Frame(self.issue_window)
        top_frame.grid(row=0, column=0, sticky=W + E)

        label = Label(top_frame, text="Select Report: ")
        label.grid(row=0, column=0, pady=10, padx=10)

        rpts = [resp.title() for resp in REPORTS]
        all_reports = ["All Reports"] + rpts

        menu = OptionMenu(
            top_frame,
            self.reports,
            *all_reports,
            command=self.process_issue_option,
        )
        menu.grid(row=0, column=1, pady=10, padx=10)

        self.grouping = StringVar()
        self.grouping.set("Error")

        label = Label(top_frame, text="Group Report by: ")
        label.grid(row=0, column=2, pady=10, padx=10)

        menu = OptionMenu(
            top_frame,
            self.grouping,
            "Error",
            "Member",
            command=self.process_issue_option,
        )
        menu.grid(row=0, column=3, pady=10, padx=10)

        txt = "Analysis..."
        top_group = LabelFrame(self.issue_window, text=txt, pady=5, padx=5)
        top_group.grid(row=1, column=0, columnspan=4, pady=10, padx=10, sticky=NSEW)

        self.issue_window.columnconfigure(0, weight=1)
        self.issue_window.rowconfigure(1, weight=1)
        top_group.columnconfigure(0, weight=1)
        top_group.rowconfigure(0, weight=1)

        self.issue_text = scrolledtext.ScrolledText(top_group, width=100, height=40)
        self.issue_text.grid(row=0, column=0, sticky=NSEW)

        self.issue_window.protocol("WM_DELETE_WINDOW", self.close_issue_window)

    def close_issue_window(self):
        """Close issue window."""
        self.issue_window.destroy()
        self.issue_window = None

    def process_issue_option(self, _):
        """Process an option selection."""
        report = self.reports.get()
        mode = self.grouping.get()

        self.issue_text.config(state=NORMAL)
        self.notify_text.config(state=NORMAL)
        self.issue_text.delete("1.0", END)

        if report == "All Reports":
            report = None
        else:
            report = report.lower()

        if mode == "Error":
            output = self.issues.print_by_error(report)
        else:
            output = self.issues.print_by_name(report)

        self.issue_text.insert(END, output)

        self.master.update_idletasks()
        self.master.after(AFTER, self.set_normal)

    def set_buttons(self, status):
        """Change button state."""
        self.button_analyse.config(state=status)
        self.button_backup.config(state=status)

        for menu, item in self.menus:
            menu.entryconfig(item, state=status)

        menu, item = self.menu_fixit

        if status == NORMAL:
            self.notify_text.config(state=DISABLED)
            length = len(self.scm.fixable)
            if length == 0:
                self.button_fixit.config(state=DISABLED)
                menu.entryconfig(item, state=DISABLED)
                return
            if self.issue_window:
                self.issue_text.config(state=DISABLED)
        else:
            self.notify_text.config(state=NORMAL)

        self.button_fixit.config(state=status)
        menu.entryconfig(item, state=status)

    def set_normal(self):
        """Set GUI to normal state after processing."""
        self.set_buttons(NORMAL)

    def new_times(self):
        """Add new time and process."""
        self.process_records(True)

    def process_records(self, newtimes=False):
        """Process Records."""
        if self.thread:
            return  # already running

        if self.scm_init() is False:
            return

        self.thread = RecordThread(self, newtimes).start()


class AnalysisThread(threading.Thread):
    """Thread to run analysis."""

    def __init__(self, gui, archive):
        """Initialise."""
        threading.Thread.__init__(self)
        self.gui = gui
        self.scm = gui.scm
        self.archive = archive

    def run(self):
        """Run analyser."""
        self.gui.set_buttons(DISABLED)

        if self.gui.issue_window:
            self.gui.issue_text.config(state=NORMAL)
            self.gui.issue_text.delete("1.0", END)

        self.gui.notify_text.delete("1.0", END)

        if self.scm.get_config_file() is False:
            messagebox.showerror("Error", "Error in config file.")
            self.gui.set_buttons(NORMAL)
            return

        if self.archive:

            home = str(Path.home())
            backup = os.path.join(home, CONFIG_DIR, BACKUP_DIR)

            dir_opt = {}
            dir_opt["initialdir"] = backup
            dir_opt["mustexist"] = True
            dir_opt["parent"] = self.gui.master

            where = filedialog.askdirectory(**dir_opt)
            if wrap(None, self.scm.decrypt, where) is False:
                messagebox.showerror("Error", f"Cannot read from archive: {where}")
                self.gui.master.after(AFTER, self.gui.set_normal)
                self.gui.thread = None
                return
        else:
            if wrap(None, self.scm.get_data, False) is False:
                messagebox.showerror("Analysis", "Failed to read data")
                self.gui.master.after(AFTER, self.gui.set_normal)
                self.gui.thread = None
                return

        if wrap(10, self.scm.linkage) is False:
            self.gui.master.after(AFTER, self.gui.set_normal)
            self.gui.thread = None
            return

        if wrap(10, self.scm.analyse) is False:
            self.gui.master.after(AFTER, self.gui.set_normal)
            self.gui.thread = None
            return

        self.gui.gotdata = True

        if self.gui.issue_window is None:
            self.gui.create_issue_window()

        debug("Analyse returned - creating result window", 8)

        output = self.gui.issues.print_by_error(None)
        result = self.scm.print_summary()

        self.gui.notify_text.insert(END, result)
        self.gui.notify_text.see(END)
        self.gui.issue_text.insert(END, output)

        self.gui.master.update_idletasks()
        self.gui.issue_window.lift()
        self.gui.master.after(AFTER, self.gui.set_normal)

        self.gui.thread = None

        debug("Analyse Thread complete, result posted", 8)

        return


class BackupThread(threading.Thread):
    """Thread to run Backuo."""

    def __init__(self, gui):
        """Initialise."""
        threading.Thread.__init__(self)
        self.gui = gui
        self.scm = gui.scm

    def run(self):
        """Run analyser."""
        self.gui.set_buttons(DISABLED)
        self.gui.notify_text.delete("1.0", END)

        if wrap(None, self.scm.backup_data):
            output = self.scm.print_summary(backup=True)
            self.gui.notify_text.insert(END, output)
            self.gui.notify_text.insert(END, "Backup Complete.")
            self.gui.notify_text.see(END)
        else:
            messagebox.showerror("Error", "Backup failure")

        self.gui.master.update_idletasks()
        self.gui.master.after(AFTER, self.gui.set_normal)

        self.gui.thread = None


class FacebookThread(threading.Thread):
    """Thread for Facebook."""

    def __init__(self, gui):
        """Initialise."""
        threading.Thread.__init__(self)
        self.gui = gui
        self.scm = gui.scm

    def run(self):
        """Process a Facebook report."""

        fbook = Facebook()
        if wrap(None, fbook.read_data, self.scm) is False:
            messagebox.showerror("Error", "Could not read facebook files")
            self.gui.report_text.config(state=DISABLED)
            self.gui.notify_text.config(state=DISABLED)
            return

        wrap(None, fbook.analyse)
        output = wrap(None, fbook.print_errors)

        if self.gui.prep_report(True, False) is False:
            return

        wrap(None, fbook.delete)
        del fbook
        self.gui.report_text.insert(END, output)
        self.gui.report_text.config(state=DISABLED)
        self.gui.notify_text.config(state=DISABLED)
        self.gui.report_window.lift()


class RecordThread(threading.Thread):
    """Thread for Records."""

    def __init__(self, gui, newtimes):
        """Initialise."""
        threading.Thread.__init__(self)
        self.gui = gui
        self.scm = gui.scm
        self.newtimes = newtimes

    def run(self):
        """Process Records."""
        self.gui.notify_text.config(state=NORMAL)
        self.gui.notify_text.delete("1.0", END)

        record = Records(self.scm)

        output = wrap(None, record.read_baseline)
        if output is False:
            self.gui.notify_text.config(state=DISABLED)
            messagebox.showerror("Error", "Cannot read records baseline")
            return

        dir_opt = {}
        dir_opt["title"] = "Locate new swim times CSV file from SCM"
        dir_opt["parent"] = self.gui.master
        dir_opt["defaultextension"] = ".csv"

        if self.newtimes:
            filename = filedialog.askopenfilename(**dir_opt)
            output = wrap(None, record.read_newtimes, filename)
            if output is False:
                self.gui.notify_text.config(state=DISABLED)
                messagebox.showerror("Error", "Cannot read new swim times")
                return

        output = wrap(None, record.create_html)
        if output:
            webbrowser.open_new(output)
        else:
            self.gui.notify_text.config(state=DISABLED)
            messagebox.showerror("Error", "Cannot Create HTML")
            return

        self.gui.notify_text.config(state=DISABLED)


class SwimEnglandThread(threading.Thread):
    """Thread for SwimEngland."""

    def __init__(self, gui):
        """Initialise."""
        threading.Thread.__init__(self)
        self.gui = gui
        self.scm = gui.scm

    def run(self):
        """Process a Swim England online report."""
        output = wrap(None, self.scm.se_check)

        if self.gui.prep_report(True, False) is False:
            return

        self.gui.report_text.insert(END, output)
        self.gui.report_text.config(state=DISABLED)
        self.gui.notify_text.config(state=DISABLED)
        self.gui.report_window.lift()


class UpdateThread(threading.Thread):
    """Thread to run Backuo."""

    def __init__(self, gui):
        """Initialise."""
        threading.Thread.__init__(self)
        self.gui = gui
        self.scm = gui.scm

    def run(self):
        """Run analyser."""
        self.gui.set_buttons(DISABLED)
        self.gui.notify_text.delete("1.0", END)

        wrap(None, self.scm.update)

        self.gui.notify_text.see(END)

        self.gui.master.update_idletasks()
        self.gui.master.after(AFTER, self.gui.set_normal)

        self.gui.thread = None


class SearchThread(threading.Thread):
    """Thread to run Search fix."""

    def __init__(self, gui):
        """Initialise."""
        threading.Thread.__init__(self)
        self.gui = gui
        self.scm = gui.scm

    def run(self):
        """Run analyser."""

        self.gui.set_buttons(DISABLED)

        if wrap(None, self.scm.fix_search) is False:
            messagebox.showerror("Error", "fix search error")
            self.gui.set_buttons(NORMAL)
            return

        self.gui.set_buttons(NORMAL)


class Edit(Frame):  # pylint: disable=too-many-ancestors
    """Class to edit a frame."""

    def __init__(self, parent, filename, scm, gui):
        """Initialise."""
        Frame.__init__(self, parent)
        self.parent = parent
        self.file = filename
        self.scm = scm
        self.gui = gui

        self.edit_win = Toplevel(self.master)
        self.edit_win.title("SCM Helper - Edit Config")

        top_frame = Frame(self.edit_win)
        top_frame.grid(row=0, column=0, sticky=W + E)

        abtn = Button(top_frame, text="Save", command=self.save_command)
        abtn.grid(row=0, column=0, pady=10, padx=10)

        cbtn = Button(top_frame, text="Close", command=self.on_exit)
        cbtn.grid(row=0, column=1, pady=10, padx=10)

        top_group = LabelFrame(self.edit_win, text=filename, pady=5, padx=5)
        top_group.grid(row=1, column=0, columnspan=2, pady=10, padx=10, sticky=NSEW)

        self.edit_win.columnconfigure(0, weight=1)
        self.edit_win.rowconfigure(1, weight=1)

        top_group.columnconfigure(0, weight=1)
        top_group.rowconfigure(0, weight=1)

        self.text_pad = scrolledtext.ScrolledText(top_group, width=80, height=40)
        self.text_pad.grid(row=0, column=0, sticky=NSEW)

        with open(self.file, FILE_READ) as file:
            contents = file.read()
            self.text_pad.insert("1.0", contents)
            file.close()

        self.text_pad.edit_modified(False)

        self.edit_win.protocol("WM_DELETE_WINDOW", self.on_exit)

    def on_exit(self):
        """Close."""
        if self.text_pad.edit_modified():
            title = "SCM-Helper: Save Config"
            msg = "Save Config?"
            resp = messagebox.askyesno(title, msg, parent=self.edit_win)
            if resp:
                if self.save_command() is False:
                    return
        self.edit_win.destroy()

    def save_command(self):
        """Save."""
        with open(self.file, FILE_WRITE) as file:
            # slice off the last character from get, as an extra return is added
            data = self.text_pad.get("1.0", END)
            file.write(data)
            file.close()

        self.gui.notify_text.config(state=NORMAL)

        if self.scm.get_config_file() is False:
            msg = "Error in config file - see status window for details."
            messagebox.showerror("Error", msg, parent=self.parent)
            self.gui.notify_text.config(state=DISABLED)
            return False

        self.text_pad.edit_modified(False)
        self.gui.notify_text.config(state=DISABLED)
        return True


def xlicense():
    """Show License."""
    messagebox.showinfo("License", LICENSE)


def xabout():
    """About message."""
    home = str(Path.home())
    cfg = os.path.join(home, CONFIG_DIR)

    msg = "SCM Helper by Colin Robbins.\n\n"
    msg += f"Version: {VERSION}.\n\n"
    msg += f"Config directory: {cfg}"

    messagebox.showinfo("About", msg)


def xhelp():
    """|Help message."""
    webbrowser.open_new(HELPURL)


def wrap_trace():
    """Give as many error details as possible."""
    tback = sys.exc_info()[2]
    while 1:
        if not tback.tb_next:
            break
        tback = tback.tb_next
    stack = []
    xframe = tback.tb_frame
    while xframe:
        stack.append(xframe)
        xframe = xframe.f_back
    stack.reverse()
    for frame in stack:
        for key, value in frame.f_locals.items():
            # Print likely to cause error itself, but should get enough out of it...
            try:
                debug(f"   {key}: {value.name}", 1)
            # pylint: disable=bare-except
            except:  # noqa: E722
                continue


def wrap(xtime, func, arg=None):
    """Catch programming logic errors."""
    try:
        if xtime is None:
            if arg is not None:
                return func(arg)
            return func()
        if arg is not None:
            return func_timeout(xtime, func, args=[arg])
        return func_timeout(xtime, func)

    except FunctionTimedOut:
        errmsg = traceback.format_exc(10)
        debug(errmsg, 1)
        nowtime = datetime.now().time()
        msg = f"{nowtime}: Abandon {func.__name__} due to timeout ({xtime} secs)"
        wrap_trace()
        messagebox.showerror("Error", msg)
        return False

    except (
        AssertionError,
        AttributeError,
        LookupError,
        NameError,
        TypeError,
        ValueError,
    ) as err:
        errmsg = f"\nSCM Helper Version: {VERSION}.\n"
        errmsg += traceback.format_exc(10)
        debug(errmsg, 0)
        msg = f"Internal SCM Helper Error:\n{err}\nPlease log an issue on github.\n"
        wrap_trace()
        messagebox.showerror("Error", msg)
        return False



"""Encryption stuff."""
import base64
import getpass
import json
import os.path
from datetime import date
from pathlib import Path

from cryptography.fernet import Fernet, InvalidToken
from cryptography.hazmat.backends import default_backend
from cryptography.hazmat.primitives import hashes
from cryptography.hazmat.primitives.kdf.pbkdf2 import PBKDF2HMAC

from scm_helper.config import BACKUP_DIR, CONFIG_DIR
from scm_helper.notify import interact, notify

WRITE_BINARY = "wb"
READ_BINARY = "rb"


class Crypto:
    """Encryption class."""

    def __init__(self, salt, password):
        """Initialise."""
        self.salt = salt
        if password is None:
            self.__password = getpass.getpass("Enter SCM helper password: ")
        else:
            self.__password = password

        self.__key = self.get_encryption_key(self.__password)

    def encrypt_file(self, filename, data):
        """Encrypt file."""
        try:
            fernet = Fernet(self.__key)

            encrypted_data = fernet.encrypt(data)
            with open(filename, WRITE_BINARY) as file:
                file.write(encrypted_data)

            notify(f"Encrypted {filename}\n")
            return True

        except OSError as error:
            notify(f"Cannot open/write file: {error}\n")
            return False
        except InvalidToken:
            notify("Cannot encrypt file - token error?\n")
            return False

    def encrypt_backup(self, name, data):
        """Encrypt file."""
        try:
            today = date.today()
            home = str(Path.home())

            backup = os.path.join(home, CONFIG_DIR, BACKUP_DIR)
            directory = os.path.join(home, CONFIG_DIR, BACKUP_DIR, f"{today}")

            if os.path.exists(backup) is False:
                os.mkdir(backup)

            if os.path.exists(directory) is False:
                os.mkdir(directory)

            filename = os.path.join(directory, f"{name}.enc")

            return self.encrypt_file(filename, data.encode("utf-8"))

        except OSError as error:
            notify(f"Cannot open/write file: {error}\n")
            return False

    def decrypt_file(self, filename):
        """Decrypt file."""
        try:
            fernet = Fernet(self.__key)

            with open(filename, READ_BINARY) as file:
                data = file.read()
            file.close()

            decrypted = fernet.decrypt(data)
            return decrypted

        except OSError as error:
            notify(f"Cannot open file: {error}\n")
            return None
        except InvalidToken:
            notify("Cannot decrypt file - wrong password?\n")
            return None

    def decrypt_backup(self, name, xdate):
        """Decrypt file."""
        try:
            home = str(Path.home())
            backup = os.path.join(home, CONFIG_DIR, BACKUP_DIR)

            filename = os.path.join(backup, xdate, f"{name}.enc")

            data = self.decrypt_file(filename)
            return json.loads(data.decode())

        except OSError as error:
            notify(f"Cannot open file: {error}\n")
            return None

    def get_encryption_key(self, password):
        """Generate a Fremat password from password and salt."""
        password = password.encode()
        salt = self.salt.encode()

        kdf = PBKDF2HMAC(
            algorithm=hashes.SHA256(),
            length=32,
            salt=salt,
            iterations=100000,
            backend=default_backend(),
        )
        return base64.urlsafe_b64encode(kdf.derive(password))

    def read_key(self, filename):
        """Read API key."""
        home = str(Path.home())
        filename = os.path.join(home, CONFIG_DIR, filename)

        if not os.path.exists(filename):
            return self.get_key(filename)
        try:
            fernet = Fernet(self.__key)

            with open(filename, READ_BINARY) as file:
                data = file.read()
            file.close()

            decrypted = fernet.decrypt(data)
            return decrypted.decode()

        except OSError as error:
            notify(f"Cannot open key file: {error}\n")
            return None
        except InvalidToken:
            notify("Cannot decrypt key file - wrong password?\n")
            return None

    def get_key(self, filename):
        """Encrypt file."""
        apikey = interact("No SCM API keyfile, creating one.  Enter API key: ")

        try:
            fernet = Fernet(self.__key)

            encrypted_data = fernet.encrypt(apikey.encode("utf-8"))
            with open(filename, WRITE_BINARY) as file:
                file.write(encrypted_data)
            file.close()

            notify(f"Generated encrypted keyfile {filename}\n")
            return apikey

        except OSError as error:
            notify(f"Cannot open/write key file: {error}\n")
            return False
        except InvalidToken:
            notify("Cannot encrypt keyfile - token error?\n")
            return False

    def read_email_password(self, filename):
        """Read email password."""
        home = str(Path.home())
        filename = os.path.join(home, CONFIG_DIR, filename)

        if not os.path.exists(filename):
            return self.get_email_password(filename)
        try:
            fernet = Fernet(self.__key)

            with open(filename, READ_BINARY) as file:
                data = file.read()
            file.close()

            decrypted = fernet.decrypt(data)
            return decrypted.decode()

        except OSError as error:
            notify(f"Cannot open password file: {error}\n")
            return None
        except InvalidToken:
            notify("Cannot decrypt password file - wrong password?\n")
            return None

    def get_email_password(self, filename):
        """Encrypt Password file."""
        apikey = getpass.getpass("Enter email password: ")

        try:
            fernet = Fernet(self.__key)

            encrypted_data = fernet.encrypt(apikey.encode("utf-8"))
            with open(filename, WRITE_BINARY) as file:
                file.write(encrypted_data)
            file.close()

            notify(f"Generate password keyfile {filename}\n")
            return apikey

        except OSError as error:
            notify(f"Cannot open/write email password file: {error}\n")
            return False
        except InvalidToken:
            notify("Cannot encrypt email password - token error?\n")
            return False



"""Check SE file."""

import json
import os.path
import pickle
import time
from pathlib import Path

import selenium

# pylint: disable=unused-import   #it is used!!!
from selenium.webdriver.common.keys import Keys

from scm_helper.config import (
    C_BASE_URL,
    C_BROWSER,
    C_CHECK_URL,
    C_DEBUG_LEVEL,
    C_SELENIUM,
    C_SWIM_ENGLAND,
    C_TEST_ID,
    C_WEB_DRIVER,
    CONFIG_DIR,
    EXCEPTION_SE_HIDDEN,
    EXCEPTION_SE_NAME,
    get_config,
)
from scm_helper.issue import debug
from scm_helper.notify import interact_yesno, notify

SE_COOKIES = "se_cookies.json"
FB_COOKIES = "fb_cookies.json.enc"
FILE_WRITE = "w"
FILE_READ = "r"

MEMBERS_SUFFIX = "members/"
SCROLL_PAUSE_TIME = 2

FACEBOOK = "https://www.facebook.com"

# M_XPATH = '//*[@id="groupsMemberSection_all_members"]'
# M_ENTRY = "//ul//div/div[2]/div/div[2]/div[1]"
# 05/07/2020 New Facebook GUI
M_XPATH = '//*[@id="mount_0_0"]'
M_ENTRY = "//div/div/div[2]/div[1]/div/div/div[1]/span/span/div/"
M_ELEMENTS = M_XPATH + M_ENTRY + "/a"
M_ELEMENTS2 = M_XPATH + M_ENTRY + "/span"


GENDER = {"M": "Male", "F": "Female"}


def se_check(scm, members):
    """Check members against the SE database."""

    home = str(Path.home())
    cookiefile = os.path.join(home, CONFIG_DIR, SE_COOKIES)

    base_url = get_config(scm, C_SWIM_ENGLAND, C_BASE_URL)
    if base_url is False:
        notify("Swim England config missing\n")
        return None

    notify(f"Opening browser for {base_url}...\n")

    browser = start_browser(scm)

    if browser is None:
        return None

    read_cookies(browser, cookiefile, base_url, None)

    check_url = get_config(scm, C_SWIM_ENGLAND, C_CHECK_URL)
    test_id_url = get_config(scm, C_SWIM_ENGLAND, C_TEST_ID)

    browser.get(f"{check_url}{test_id_url}")
    try:
        browser.find_element_by_xpath("//table[1]/tbody/tr[2]/td[2]")

    except selenium.common.exceptions.NoSuchElementException:
        msg = """Please solve the 'I am not a robot', and then press enter here.
If you cnnot solve the capture try deleting the file 'se_cookies.json'
in the config directory.
"""
        interact_yesno(msg)
        write_cookies(browser, cookiefile, None)

    res = ""
    for member in members:
        if member.is_active is False:
            continue
        if member.asa_number is None:
            continue

        url = f"{check_url}{member.asa_number}"
        browser.get(url)

        res += check_member(browser, member)

    browser.close()

    return res


def fb_read_url(scm, url):
    """Read Facebook Group members."""

    users = []

    home = str(Path.home())
    cookiefile = os.path.join(home, CONFIG_DIR, FB_COOKIES)

    notify(f"Opening browser for {url}...\n")

    url += MEMBERS_SUFFIX

    browser = start_browser(scm)

    if browser is None:
        return None

    read_cookies(browser, cookiefile, FACEBOOK, scm)

    browser.get(url)
    try:
        browser.find_element_by_xpath(M_XPATH)
    except selenium.common.exceptions.NoSuchElementException:
        interact_yesno("Please logon to Facebook and then press enter here.")
        write_cookies(browser, cookiefile, scm)
        browser.get(url)

    debug_config = scm.config(C_DEBUG_LEVEL)
    if debug_config > 1:
        interact_yesno(
            "Navigate to members page (if not there) - then Click to Continue"
        )

    scroll(browser)

    count = 0
    for path in [M_ELEMENTS, M_ELEMENTS2]:
        elements = browser.find_elements_by_xpath(path)
        for element in elements:
            if element.text not in users:
                count += 1
                users.append(element.text)

    notify(f"Read {count} names from {url}\n")
    browser.close()

    return users


def scroll(browser):
    """Scroll to end of page."""

    # Get scroll height
    last_height = browser.execute_script("return document.body.scrollHeight")

    notify("Scrolling")

    while True:
        # Scroll down to bottom
        browser.execute_script("window.scrollTo(0, document.body.scrollHeight);")

        # Wait to load page
        time.sleep(SCROLL_PAUSE_TIME)
        notify(".")

        # Calculate new scroll height and compare with last scroll height
        new_height = browser.execute_script("return document.body.scrollHeight")
        if new_height == last_height:
            notify("Done\n")
            return
        last_height = new_height


def check_member(browser, member):
    """Check a member."""

    notify(f"Checking {member.name}...\n")

    try:
        name = browser.find_element_by_xpath("//table[1]/tbody/tr[2]/td[2]").text
        knownas = browser.find_element_by_xpath("//table[1]/tbody/tr[2]/td[4]").text
        gender = browser.find_element_by_xpath("//table[1]/tbody/tr[3]/td[4]").text
        current = browser.find_element_by_xpath("//table[2]/tbody/tr[2]/td[4]").text
        category = browser.find_element_by_xpath("//table[2]/tbody/tr[3]/td[4]").text

    except selenium.common.exceptions.NoSuchElementException:
        if member.print_exception(EXCEPTION_SE_HIDDEN) is False:
            debug(f"SE Exception ignored: {member.name}", 7)
            return ""

        res = f"\n{member.name} ({member.asa_number}) does not exist in SE database.\n"
        return res

    res = ""

    if member.print_exception(EXCEPTION_SE_NAME) is True:
        if name.lower() != member.name.lower():
            res += f"   Name: SCM-> {member.name}, SE-> {name}\n"

        if knownas and (member.knownas_only != knownas):
            firstname = name.split(" ")
            if knownas != firstname[0]:  # in SE they are the same if no knownas
                res += f"   Knownas: SCM-> {member.knownas_only}, SE-> {knownas}\n"

    if member.gender and (gender != GENDER[member.gender]):
        res += f"   Gender: SCM-> {member.name}, SE-> {gender}\n"

    mycat = f"SE Category {member.asa_category}"
    if category != mycat:
        res += f"   Category: SCM-> {mycat}, SE-> {category}\n"

    if current != "Current":
        res += "   Not current\n"

    if res:
        res = f"\n{member.name} ({member.asa_number}) mismatch:\n" + res

    return res


def start_browser(scm):
    """Start Browser."""

    web_driver = get_config(scm, C_SELENIUM, C_WEB_DRIVER)
    client = get_config(scm, C_SELENIUM, C_BROWSER)

    if client is False:
        notify("Selenium config missing\n")
        return None

    browser = getattr(selenium.webdriver, client)

    try:
        return browser(web_driver)

    except selenium.common.exceptions.WebDriverException as error:
        notify(f"Failed to open {client} browser with: {web_driver}\n{error}\n")
        return None


def read_cookies(browser, cookiefile, url, scm):
    """Read cookies."""

    if os.path.isfile(cookiefile):
        browser.get(url)

        if scm:
            data = scm.crypto.decrypt_file(cookiefile)
            if data:
                cookies = pickle.loads(data)
        else:
            try:
                with open(cookiefile, FILE_READ) as file:
                    data = file.read()
                    cookies = json.loads(data)

            except EnvironmentError as error:
                notify(f"Failed to read {cookiefile}\n{error}\n")

        for cookie in cookies:
            if "expiry" in cookie:
                del cookie["expiry"]
            browser.add_cookie(cookie)


def write_cookies(browser, cookiefile, scm):
    """Write cookies."""

    cookies = browser.get_cookies()

    if cookies:
        if scm:
            data = pickle.dumps(cookies)
            scm.crypto.encrypt_file(cookiefile, data)
            return

        try:
            with open(cookiefile, FILE_WRITE) as file:
                opt = json.dumps(cookies)
                file.write(opt)

        except EnvironmentError as error:
            notify(f"Failed to write {cookiefile}\n{error}\n")



"""SCM Members."""
import datetime
import re

from scm_helper.coach import analyse_coach
from scm_helper.conduct import check_conduct
from scm_helper.config import (
    A_ACTIVE,
    A_ASA_CATEGORY,
    A_ASA_NUMBER,
    A_DOB,
    A_FIRSTNAME,
    A_GUID,
    A_ISCOACH,
    A_ISPARENT,
    A_ISVOLUNTEER,
    A_KNOWNAS,
    A_LASTNAME,
    A_PARENTS,
    A_USERNAME,
    C_ALIGN_QUARTER,
    C_CONFIRMATION,
    C_DBS,
    C_EXPIRY,
    C_GRACE,
    C_GROUP,
    C_GROUPS,
    C_IGNORE,
    C_IGNORE_COACH,
    C_IGNORE_COMMITTEE,
    C_IGNORE_UNKNOWN,
    C_INACTIVE,
    C_JOBTITLE,
    C_LISTS,
    C_MEMBERS,
    C_NAME,
    C_NEWSTARTER,
    C_PRIORITY,
    C_TIME,
    C_TYPES,
    CTYPE_COACH,
    CTYPE_COMMITTEE,
    CTYPE_PARENT,
    CTYPE_POLO,
    CTYPE_SWIMMER,
    CTYPE_SYNCHRO,
    CTYPE_VOLUNTEER,
    EXCEPTION_NODBS,
    EXCEPTION_NOEMAIL,
    EXCEPTION_NOSAFEGUARD,
    PRINT_DATE_FORMAT,
    SCM_DATE_FORMAT,
    get_config,
)
from scm_helper.entity import Entity, get_date, print_all
from scm_helper.issue import (
    E_CONFIRMATION_EXPIRED,
    E_DATE,
    E_DBS_EXPIRED,
    E_EMAIL_SPACE,
    E_INACTIVE_TOOLONG,
    E_JOB,
    E_NAME_CAPITAL,
    E_NO_DBS,
    E_NO_EMAIL,
    E_NO_JOB,
    E_NO_LEAVE_DATE,
    E_NO_SAFEGUARD,
    E_NOT_CONFIRMED,
    E_OWNPARENT,
    E_SAFEGUARD_EXPIRED,
    E_TOO_OLD,
    E_TOO_YOUNG,
    E_TYPE_GROUP,
    E_UNKNOWN,
    debug,
    debug_trace,
    issue,
)
from scm_helper.notify import notify
from scm_helper.parent import analyse_parent
from scm_helper.swimmer import analyse_swimmer

FACEBOOK_RE = re.compile(r"Facebook: *([a-zA-Z\- ]+)")
API_RE = re.compile(r"API:.+")  # The whole line
API_TEXT_RE = re.compile(r"API: *[a-zA-Z ]+")  # excluding date
DATE_RE = re.compile(r"\d\d-\d\d-\d\d\d\d")  # date
DATE2_RE = re.compile(r"\d\d/\d\d/\d\d\d\d")  # date

A_DATELEFT = "DateLeft"
A_DBS_RENEWAL_DATE = "DBSRenewalDate"
A_SAFEGUARDING_RENEWAL_DATE = "SafeguardingRenewalDate"
A_SWIMMERS = "Swimmers"


class Member(Entity):
    """A member."""

    # pylint: disable=too-many-instance-attributes
    # Need them all!
    # pylint: disable=too-many-public-methods
    # Need them all!

    def __init__(self, entity, scm, url):
        """Initialise."""
        super().__init__(entity, scm, url)
        self._parents = []
        self._swimmers = []
        self._sessions = []
        self._coach_sessions = []
        self._restricted = []
        self.facebook = []
        self.groups = []
        self._conduct = []
        self.ignore_errors = []
        self._in_ignore_swimmer = False
        self._in_ignore_group = False
        self._first_group = None
        self._lastseen = None
        self._in_coach_role = False
        self._no_session_ok = False

        self._dob = None
        self._date_joined = None
        self._last_modified = None
        self._confirmed_date = None
        self._last_login = None

        self.set_dates()
        self.get_notes()

    def get_notes(self):
        """Extract Facebook name from Notes."""
        notes = self.notes
        if notes is None:
            return

        note = FACEBOOK_RE.findall(notes)
        if note:
            for facebook in note:
                facebook = facebook.strip()
                self.facebook.append(facebook)
                debug(f"Found Facebook name in notes '{facebook}'", 8)

        note = API_RE.findall(notes)
        if note is None:
            return
        for api in note:
            exclusion = API_TEXT_RE.search(api)
            expiry = DATE_RE.search(api)
            when = self.scm.today
            gotdate = False
            if expiry:
                date = expiry.group(0)
                when = get_date(date, "%d-%m-%Y")
                gotdate = True
            else:
                expiry = DATE2_RE.search(api)
                if expiry:
                    date = expiry.group(0)
                    when = get_date(date, "%d/%m/%Y")
                    gotdate = True
            if when:
                excl = exclusion.group(0).strip()
                if (self.scm.today - when).days <= 0:
                    self.ignore_errors.append(excl)
                    debug(f"Found API token in notes {api}", 8)
                else:
                    debug(f"Token expired {api}", 8)
            elif gotdate:
                issue(self, E_DATE, f"Notes: {api}")

    def print_notes(self):
        """Print notes (if any)."""
        msg = f"{self.name}:\n"
        found = False
        for note in self.facebook:
            found = True
            msg += f"   Facebook: {note}\n"
        for note in self.ignore_errors:
            found = True
            msg += f"   {note}\n"
        if found:
            return msg
        return ""

    def linkage_parent(self, members):
        """Link parents."""
        for parent in self.data[A_PARENTS]:
            guid = members.by_guid[parent[A_GUID]]

            if guid == self.guid:
                issue(self, E_OWNPARENT)

            self._parents.append(guid)

    def linkage_swimmer(self, members):
        """Link swimmers."""
        for swimmer in self.data[A_SWIMMERS]:
            guid = members.by_guid[swimmer[A_GUID]]
            self._swimmers.append(guid)

    def linkage_restrictions(self):
        """Link restrictinos."""
        for session in self.session_restrictions:
            guid = self.scm.sessions.by_guid[session[A_GUID]]
            self._restricted.append(guid)

    def linkage2(self):
        """Link parents to swimmers."""
        # Hack - work around API issue.
        # if a parent, make sure swimmers are linked back
        # pylint: disable=protected-access
        for swimmer in self._swimmers:
            if len(swimmer._parents) == 0:
                debug(f"Found swimmer - API error - recovered {swimmer.name}", 7)
                swimmer._parents.append(self)

    def linkage(self, members):
        """Link parents and swimmers."""
        if self.data[A_PARENTS]:
            self.linkage_parent(members)

        if self.data[A_SWIMMERS]:
            self.linkage_swimmer(members)

        if self.session_restrictions:
            self.linkage_restrictions()

        self._first_group = self.set_first_group()

    def find_session_substr(self, substring):
        """Find a session containing substring."""
        for session in self.sessions:
            if session.name.find(substring) >= 0:
                return True
        return False

    def find_group(self, name):
        """Find a group containing string."""
        for group in self.groups:
            if name == group.name:
                return True
        return False

    def find_conduct(self, find):
        """Find a group containing string."""
        for codes in self._conduct:
            if find == codes:
                return True
        return False

    def get_conduct_name(self, find):
        """get a code of conduct."""
        for code in self._conduct:
            if find == code.name:
                return code
        return None

    def check_email(self):
        """Check email."""
        email = self.email
        if email is None:
            if self.print_exception(EXCEPTION_NOEMAIL) is False:
                return
            issue(self, E_NO_EMAIL, f"{self.first_group}")
        else:
            space = re.search(" ", email)
            if space:
                issue(self, E_EMAIL_SPACE, f"{email}")

    def check_dbs(self, xtype):
        """Check DBS and Safeguarding."""
        if self.print_exception(EXCEPTION_NODBS) is False:
            debug(f"DBS Exception ignored: {self.name}", 7)
            return

        dbs_date = self.set_date(A_DBS_RENEWAL_DATE)
        safe_date = self.set_date(A_SAFEGUARDING_RENEWAL_DATE)
        notice = get_config(self.scm, C_MEMBERS, C_DBS, C_EXPIRY)

        if dbs_date:
            days = (dbs_date - self.scm.today).days
            if days < 0:
                dbs_date_str = dbs_date.strftime(PRINT_DATE_FORMAT)
                issue(self, E_DBS_EXPIRED, f"{xtype}, expired {dbs_date_str}")
            elif days < notice:
                dbs_date_str = dbs_date.strftime(PRINT_DATE_FORMAT)
                issue(self, E_DBS_EXPIRED, f"{xtype}, expires {dbs_date_str}")
        else:
            issue(self, E_NO_DBS, f"{xtype}")

        if self.print_exception(EXCEPTION_NOSAFEGUARD) is False:
            debug(f"Safeguard Exception ignored: {self.name}", 7)
            return

        if safe_date:
            if (safe_date - self.scm.today).days < 0:
                safe_date_str = safe_date.strftime(PRINT_DATE_FORMAT)
                issue(self, E_SAFEGUARD_EXPIRED, f"{xtype}, expired {safe_date_str}")
            elif (safe_date - self.scm.today).days < notice:
                safe_date_str = safe_date.strftime(PRINT_DATE_FORMAT)
                issue(self, E_SAFEGUARD_EXPIRED, f"{xtype}, expires {safe_date_str}")
        else:
            issue(self, E_NO_SAFEGUARD, f"{xtype}")

    def check_inactive(self):
        """Check an inactive member."""
        lastmod = self.last_modified_date
        if lastmod:
            gap = (self.scm.today - lastmod).days
            inactive = get_config(self.scm, C_MEMBERS, C_INACTIVE, C_TIME)
            if inactive and (gap > inactive):
                when = lastmod.strftime(PRINT_DATE_FORMAT)
                issue(self, E_INACTIVE_TOOLONG, f"Last Modified: {when}")

        if self.check_attribute(A_DATELEFT):
            return

        if lastmod:
            issue(self, E_NO_LEAVE_DATE, "fixable", -1)
            fix = {}
            fix[A_DATELEFT] = lastmod.strftime(SCM_DATE_FORMAT)
            self.fixit(fix, f"Add dataleft = {fix[A_DATELEFT]}")
            return

        issue(self, E_NO_LEAVE_DATE, "", -1)

    def _list_add(self, err):
        """Add a member to a confirmation list."""
        msg = "Not confirmed"
        if err == E_CONFIRMATION_EXPIRED:
            msg = "Confirmation expired"

        xtype = " (Other)"
        if self.is_parent:
            xtype = " (Parent)"
        if self.is_polo:
            xtype = " (Water Polo)"
            cfg = get_config(self.scm, C_TYPES, CTYPE_POLO, C_NAME)
            if cfg:
                xtype = f" ({cfg})"
        if self.is_synchro:
            xtype = " (Synchro)"
            cfg = get_config(self.scm, C_TYPES, CTYPE_SYNCHRO, C_NAME)
            if cfg:
                xtype = f" ({cfg})"
        if self.is_swimmer:
            xtype = " (Swimmer)"

        msg += xtype

        self.scm.lists.add(msg, self)

    def check_confirmation(self):
        """Check confimation status."""
        expiry = get_config(self.scm, C_MEMBERS, C_CONFIRMATION, C_EXPIRY)
        align = get_config(self.scm, C_MEMBERS, C_CONFIRMATION, C_ALIGN_QUARTER)
        xlist = get_config(self.scm, C_LISTS, C_CONFIRMATION)
        q_offset = 0

        if align:
            q_offset = self.scm.q_offset

        if self.confirmed_date:
            gap = (self.scm.today - self.confirmed_date).days
            if gap > (expiry + q_offset):
                issue(self, E_CONFIRMATION_EXPIRED, f"{self.first_group}")
                self.scm.members.count_not_confirmed += 1
                if xlist:
                    self._list_add(E_CONFIRMATION_EXPIRED)
        else:
            issue(self, E_NOT_CONFIRMED, f"{self.first_group}")
            self.scm.members.count_not_confirmed += 1
            if xlist:
                self._list_add(E_NOT_CONFIRMED)

    def check_name(self):
        """Check capitilisation of name."""
        firstname = self.data[A_FIRSTNAME]
        lastname = self.data[A_LASTNAME]
        knownas = self.data[A_KNOWNAS]
        ka_upper = True

        if knownas:
            ka_upper = knownas[0].isupper()

        fn_upper = firstname[0].isupper()
        ln_upper = lastname[0].isupper()

        if fn_upper and ln_upper and ka_upper:
            return

        if (fn_upper is False) or (ln_upper is False):
            issue(self, E_NAME_CAPITAL, "fixable", -1)

            fix = {}
            if fn_upper is False:
                fix[A_FIRSTNAME] = firstname.title()
            if ln_upper is False:
                fix[A_LASTNAME] = lastname.title()
            if ka_upper is False:
                fix[A_KNOWNAS] = knownas.title()

            self.fixit(fix, "Capitalisation of name")
            return

        issue(self, E_NAME_CAPITAL, f"Knownas = {knownas}", -1, "fixable")
        fix = {}
        fix[A_KNOWNAS] = knownas.title()
        self.fixit(fix, f"Capitalisation of {knownas}")

    def check_type(self, xtype):
        """Check the member type check box config."""
        cfg = get_config(self.scm, C_TYPES, xtype, C_GROUPS)
        name = get_config(self.scm, C_TYPES, xtype, C_NAME)
        if name is None:
            name = xtype
        jobtitle = get_config(self.scm, C_TYPES, xtype, C_JOBTITLE)
        if cfg:
            found = False
            for group in cfg:
                if self.find_group(group):
                    found = True
                    break
            if found is False:
                if self.in_ignore_group is False:
                    if self.in_ignore_swimmer is False:
                        issue(self, E_TYPE_GROUP, name)

        if jobtitle:
            if self.jobtitle:
                return
            issue(self, E_NO_JOB, f"{name}", 0, "fixable")
            fix = {}
            fix["JobTitle"] = xtype.title()
            self.fixit(fix, f"Add jobtitle: {name}")

    @debug_trace(5)
    def analyse(self):
        """Analyse the member."""
        # pylint: disable=too-many-branches
        # pylint: disable=too-many-statements

        if self.is_active is False:
            self.check_inactive()
            return

        self.check_name()

        found = False

        if self.is_coach:
            self.check_type(CTYPE_COACH)
            analyse_coach(self)
            found = True

        if self.newstarter:
            return

        if self.is_swimmer:
            self.check_type(CTYPE_SWIMMER)
            found = True

        if self.is_synchro:
            self.check_type(CTYPE_SYNCHRO)
            found = True

        if self.is_polo:
            self.check_type(CTYPE_POLO)
            found = True

        if self.is_swimmer or self.is_polo or self.is_synchro or self.is_openwater:
            analyse_swimmer(self)

        if self.is_parent:
            self.check_type(CTYPE_PARENT)
            analyse_parent(self)
            found = True

        if self.is_volunteer:
            vol = get_config(self.scm, C_TYPES, CTYPE_VOLUNTEER, C_IGNORE_COACH)
            ctte = get_config(self.scm, C_TYPES, CTYPE_VOLUNTEER, C_IGNORE_COMMITTEE)

            if vol or ctte or self.is_coach:
                pass
            else:
                self.check_type(CTYPE_VOLUNTEER)
            found = True

        if self.is_committee_member:
            found = True
            self.check_type(CTYPE_COMMITTEE)

        if self.jobtitle:
            if (
                (self.is_volunteer is not True)
                and (self.is_committee_member is not True)
                and (self.is_coach is not True)
            ):
                cfg = get_config(self.scm, C_JOBTITLE, C_IGNORE)
                if self.jobtitle not in cfg:
                    issue(self, E_JOB, self.jobtitle)

        if found:
            if self.in_ignore_swimmer:
                return
            self.check_email()
            self.check_confirmation()  # Must come after analyse_swimmer
            check_conduct(self, self._conduct)
            return

        for group in self.groups:
            if get_config(self.scm, C_GROUPS, C_GROUP, group.name, C_IGNORE_UNKNOWN):
                continue
            issue(self, E_UNKNOWN)

    def fix_search(self):
        """fix_search_index."""
        if A_KNOWNAS in self.data:
            if self.data[A_KNOWNAS]:

                self.newdata = {}
                self.newdata[A_GUID] = self.guid
                notify(f"Deleting index for {self.name}...")

                self.newdata[A_FIRSTNAME] = "XXX"

                res = self.scm.api_write(self, False)
                if res is False:
                    notify("Hit a snag!\n")
                    return False

                notify("Recreating...")

                self.newdata[A_FIRSTNAME] = self.data[A_FIRSTNAME]

                res = self.scm.api_write(self, False)
                if res is False:
                    msg = "Hit a snag - Firstname"
                    msg2 = "deleted - oops sorry - restore manually!\n"
                    notify(f"{msg} '{self.data[A_FIRSTNAME]}' {msg2}")
                    return False

                notify("Success.\n")

        return True

    def add_group(self, group):
        """Add a group to the swimmer."""
        debug(f"Added {self.name} to {group.name}", 9)
        self.groups.append(group)

    def add_session(self, session):
        """Add swimemrs sessions."""
        self._sessions.append(session)

    def add_conduct(self, conduct):
        """Add a conduct to the swimmer."""
        self._conduct.append(conduct)

    def add_coach_session(self, session):
        """Add coaches sessions."""
        self._coach_sessions.append(session)

    def add_restrictions(self, restriction):
        """Add swimmers restricted sessions."""
        self._restricted.append(restriction)

    def set_in_coach_role(self):
        """Member of a role."""
        self._in_coach_role = True

    def set_first_group(self):
        """Print the members primary group."""
        cfg_groups = self.scm.config(C_GROUPS)
        if cfg_groups is None:
            if self.groups:
                return self.groups[0]
            return None
        group_priority = cfg_groups[C_PRIORITY]
        if self.groups and group_priority:
            for group in self.groups:
                for priority in group_priority:
                    if group.name == priority:
                        return group

            # Pick the first one w
            return self.groups[0]
        return None

    def set_lastseen(self, lastseen):
        """Set when the swimmer was last seen."""
        when = datetime.datetime.strptime(lastseen, SCM_DATE_FORMAT)
        if self._lastseen is None:
            self._lastseen = when
            return
        if when > self._lastseen:
            self._lastseen = when

    def set_confirmed(self, date):
        """Set confirmed date."""
        self._confirmed_date = date

    def set_no_session_ok(self):
        """Set no session paramet to True."""
        self._no_session_ok = True

    def set_ignore_group(self, state):
        """Set ignore group."""
        self._in_ignore_group = state

    def set_ignore_swimmer(self, state):
        """Set ignore swimmer."""
        self._in_ignore_swimmer = state

    def print_exception(self, exception):
        """Return true if the exception allowable."""
        if len(self.ignore_errors) == 0:
            return True
        for item in self.ignore_errors:
            if exception == item:
                return False
        return True

    def print_swimmer_sessions(self, grid):
        """Print sessions per swimmers."""
        if len(self.sessions) == 0:
            return None

        res = ""
        cnt = 0

        if grid:
            for session in self.scm.sessions.entities:
                if session.is_active:
                    if session in self.sessions:
                        res += "X"
                    res += ","
        else:
            for session in self.sessions:
                if cnt > 0:
                    res += "\n"
                else:
                    cnt = 1
                res += f"        {session.full_name}"

        return res

    def print_links(self):
        """Print links."""
        # pylint: disable=no-self-use
        # override if anything needs doing
        res = "You will need to restore the following manually\n"
        res += "Sessions: "
        res += print_all(self.sessions)
        res += "\nGroups: "
        res += print_all(self.groups)
        res += "\nParents: "
        res += print_all(self.parents)
        res += "\nSwimmers: "
        res += print_all(self.swimmers)
        res += "\n"

    @property
    def first_group(self):
        """Print the members primary group."""
        if self._first_group:
            return self._first_group.name
        return "No Group"

    @property
    def name(self):
        """Name."""
        firstname = self.data[A_FIRSTNAME]
        lastname = self.data[A_LASTNAME]
        return f"{firstname} {lastname}"

    @property
    def full_name(self):
        """Full name - use Knownas."""
        firstname = self.data[A_FIRSTNAME]
        lastname = self.data[A_LASTNAME]
        if A_KNOWNAS in self.data:
            if self.data[A_KNOWNAS]:
                return f"{self.data[A_KNOWNAS]} ({firstname}) {lastname}"
        return self.name

    @property
    def knownas(self):
        """Known as."""
        firstname = self.data[A_FIRSTNAME]

        if A_KNOWNAS in self.data:
            if self.data[A_KNOWNAS]:
                firstname = self.data[A_KNOWNAS]

        lastname = self.data[A_LASTNAME]
        return f"{firstname} {lastname}"

    @property
    def knownas_only(self):
        """Known as."""
        firstname = self.data[A_FIRSTNAME]
        if A_KNOWNAS in self.data:
            if self.data[A_KNOWNAS]:
                firstname = self.data[A_KNOWNAS]
        return firstname

    @property
    def newstarter(self):
        """Is the member a new stater."""
        grace = get_config(self.scm, C_MEMBERS, C_NEWSTARTER, C_GRACE)
        if grace and self._date_joined:
            if (self.scm.today - self._date_joined).days < grace:
                return True
        return False

    @property
    def age(self):
        """Calculate age."""
        myage = None
        if self._dob:
            myage = (self.scm.today - self._dob).days // 365
        return myage

    @property
    def age_eoy(self):
        """Calculate age at eof of year."""
        myage = None
        if self._dob:
            myage = (self.scm.eoy - self._dob).days // 365
        return myage

    def check_age(self, xmin, xmax, group):
        """Check swimmer within age group."""
        if self.age and (self.age < xmin):
            issue(self, E_TOO_YOUNG, f"{group}: {self.age}")
        if self.age and (self.age > xmax):
            if not self.is_coach:
                issue(self, E_TOO_OLD, f"{group}: {self.age}")

    def set_dates(self):
        """Calculate dates."""
        self._dob = self.set_date(A_DOB)
        self._date_joined = self.set_date("DateJoinedClub")
        self._last_modified = self.set_date("LastModifiedDate")
        self._confirmed_date = self.set_date("DetailsConfirmedCorrect")
        self._last_login = self.set_date("LastLoggedIn")

    def set_joined_today(self):
        """Set joined date."""
        self._date_joined = self.scm.today

    @property
    def confirmed_date(self):
        """Set confirmed date Date."""
        return self._confirmed_date

    @property
    def dob(self):
        """Set DOB."""
        return self._dob

    @property
    def date_joined(self):
        """Set joining date."""
        return self._date_joined

    @property
    def last_login(self):
        """Set last login date."""
        return self._last_login

    @property
    def last_modified_date(self):
        """Set Last Modified Date."""
        return self._last_modified

    @property
    def asa_number(self):
        """Set ASA Number."""
        return self.check_attribute(A_ASA_NUMBER)

    @property
    def asa_category(self):
        """Set ASA Number."""
        return self.check_attribute(A_ASA_CATEGORY)

    @property
    def notes(self):
        """Set Notes."""
        return self.check_attribute("Notes")

    @property
    def username(self):
        """Set Username."""
        return self.check_attribute(A_USERNAME)

    @property
    def homephone(self):
        """Set Username."""
        return self.check_attribute("HomePhone")

    @property
    def mobilephone(self):
        """Set Username."""
        return self.check_attribute("MobilePhone")

    @property
    def address(self):
        """Set Username."""
        return self.check_attribute("Address1")

    @property
    def is_active(self):
        """Is the entry active..."""
        isa = self.check_attribute(A_ACTIVE)
        if isa == "1":
            return True
        return False

    @property
    def is_swimmer(self):
        """Is it a swimmer."""
        isa = self.check_attribute("IsASwimmer")
        if isa == "1":
            return True
        return False

    @property
    def is_coach(self):
        """Is it a Coach."""
        isa = self.check_attribute(A_ISCOACH)
        if isa == "1":
            return True
        return False

    @property
    def is_parent(self):
        """Is it a swimmer."""
        isa = self.check_attribute(A_ISPARENT)
        if isa == "1":
            return True
        return False

    @property
    def is_synchro(self):
        """Is it a syncro swimmer."""
        isa = self.check_attribute("SynchronisedSwimming")
        if isa == "1":
            return True
        return False

    @property
    def is_lts(self):
        """Is it a LTS Teacher."""
        isa = self.check_attribute("LTSTeacher")
        if isa == "1":
            return True
        return False

    @property
    def is_openwater(self):
        """Is it a Open Water swimmer."""
        isa = self.check_attribute("OpenWater")
        if isa == "1":
            return True
        return False

    @property
    def is_lifesaving(self):
        """Is it a Life saver."""
        isa = self.check_attribute("LifeSaving")
        if isa == "1":
            return True
        return False

    @property
    def is_polo(self):
        """Is it a polo player."""
        isa = self.check_attribute("WaterPolo")
        if isa == "1":
            return True
        return False

    @property
    def is_committee_member(self):
        """Is it a CommitteeMember."""
        isa = self.check_attribute("CommitteeMember")
        if isa == "1":
            return True
        return False

    @property
    def is_volunteer(self):
        """Is it a Volunteer."""
        isa = self.check_attribute(A_ISVOLUNTEER)
        if isa == "1":
            return True
        return False

    @property
    def email(self):
        """Set email."""
        email = self.check_attribute("Email")
        if email:
            return email.lower()
        return None

    @property
    def gender(self):
        """Set Gender."""
        return self.check_attribute("Gender")

    @property
    def session_restrictions(self):
        """Return sessions restrictions."""
        return self.check_attribute("SessionRestrictions")

    @property
    def jobtitle(self):
        """Return job title."""
        return self.check_attribute("JobTitle")

    @property
    def lastseen_str(self):
        """Set lastseen as a string."""
        if self._lastseen:
            return self._lastseen.strftime(PRINT_DATE_FORMAT)
        return "Never"

    @property
    def lastseen(self):
        """Set lastseen."""
        return self._lastseen

    @property
    def swimmers(self):
        """Return swimmers."""
        return self._swimmers

    @property
    def parents(self):
        """Return parents."""
        return self._parents

    @property
    def sessions(self):
        """Return sessions."""
        return self._sessions

    @property
    def coach_sessions(self):
        """Return sessions."""
        return self._coach_sessions

    @property
    def restricted(self):
        """Return sessions."""
        return self._restricted

    @property
    def coach_role(self):
        """Return sessions."""
        return self._in_coach_role

    @property
    def no_session_ok(self):
        """Return true is no sessions OK."""
        return self._no_session_ok

    @property
    def in_ignore_group(self):
        """Return true is no sessions OK."""
        return self._in_ignore_group

    @property
    def in_ignore_swimmer(self):
        """Return true is no sessions OK."""
        return self._in_ignore_swimmer

    @property
    def print_groups(self):
        """Print out the groups."""
        return print_all(self.groups)
